package MIC.extaction;

import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.action.police.ActionClear;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.PathPlanning;
import clear.Geometry;
import com.google.common.collect.Iterables;

import MIC.util.ExtraActionsUtil;
import MIC.util.Agent.MIC_ATConfig;

import org.apache.commons.math3.random.MersenneTwister;
import org.apache.commons.math3.random.RandomGenerator;
import rescuecore2.config.NoSuchConfigOptionException;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.standard.entities.Area;
import rescuecore2.worldmodel.EntityID;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.util.List;

public class MICActionExtClear extends ExtAction {
    private PathPlanning pathPlanning;

    private int clearDistance;
    private int clearRad;
    private int clearRate;
    private int thresholdRest;
    private int kernelTime;

    private EntityID target;

    private int clearAreaPadding;
    private int passableWidth;
    private Action lastAction;
    private Point2D lastPosition;
    protected RandomGenerator random;


    public MICActionExtClear(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
    	
    	/***
    	 * 
    	 * Thanks a lot Okan.
    	 * Copied from RoboAKUT Extclear
    	 * 
    	 */
    	
    	
    	
    	
    	
        super(ai, wi, si, moduleManager, developData);
        this.clearDistance = si.getClearRepairDistance();
        this.clearRad = si.getClearRepairRad();
        this.clearRate = si.getClearRepairRate();
        this.thresholdRest = developData.getInteger("ActionExtClear.rest", 100);
        int randomSeed = developData.getInteger("MICActionExtClear.randomSeed", 1);

        this.target = null;

        this.clearAreaPadding = -10;
        this.passableWidth = 300;

        this.lastAction = null;
        this.lastPosition = null;
        random = new MersenneTwister(randomSeed);

        switch  (si.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
        }

    }

    @Override
    public ExtAction precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        return this;
    }

    @Override
    public ExtAction resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        preparate();
        return this;
    }

    @Override
    public ExtAction preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.pathPlanning.preparate();
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction updateInfo(MessageManager messageManager){
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.pathPlanning.updateInfo(messageManager);
        return this;
    }

    @Override
    public ExtAction setTarget(EntityID target) {
        this.target = null;
        StandardEntity entity = this.worldInfo.getEntity(target);
        if(entity != null) {
            if (entity instanceof Road) {
                this.target = target;
            } else if (entity.getStandardURN().equals(StandardEntityURN.BLOCKADE)) {
                this.target = ((Blockade) entity).getPosition();
            } else if (entity instanceof Building) {
                this.target = target;
            }
        }
        return this;
    }

    protected Vector2D rotate(Vector2D vector, double angle) {
        double rotatedX = vector.getX()*Math.cos(angle)-vector.getY()*Math.sin(angle);
        double rotatedY = vector.getX()*Math.sin(angle)+vector.getY()*Math.cos(angle);
        return new Vector2D(rotatedX, rotatedY);
    }

    protected Path2D createPathFromLines(List<Line2D> lines) {
        Path2D path = new Path2D.Float();
        if (lines.size() == 1) {
            path.moveTo(lines.get(0).getOrigin().getX(), lines.get(0).getOrigin().getY());
            path.lineTo(lines.get(0).getEndPoint().getX(), lines.get(0).getEndPoint().getY());
            return path;
        }
        path.moveTo(lines.get(0).getOrigin().getX(), lines.get(0).getOrigin().getY());
        for (int i = 1; i < lines.size(); i++) {
            Line2D line = lines.get(i);
            path.lineTo(line.getOrigin().getX(), line.getOrigin().getY());
        }
        Line2D lastLine = lines.get(lines.size()-1);
        path.lineTo(lastLine.getEndPoint().getX(), lastLine.getEndPoint().getY());
        path.closePath();
        return path;
    }

    protected List<java.awt.geom.Line2D> convertLinesToAwtLines(List<Line2D> lines) {
        List<java.awt.geom.Line2D> awtLines = new ArrayList<>(lines.size());
        for (Line2D line : lines) {
            awtLines.add(new java.awt.geom.Line2D.Double(line.getOrigin().getX(), line.getOrigin().getY(), line.getEndPoint().getX(), line.getEndPoint().getY()));
        }
        return awtLines;
    }

    protected List<java.awt.geom.Line2D> covertAreaToLines(java.awt.geom.Area area, Shape shape) {
        List<java.awt.geom.Line2D> awtLines = new ArrayList<>();
        PathIterator iter = area.getPathIterator(null);
        double prevPoint[] = null;
        while(!iter.isDone()) {
            double point[] = new double[2];
            int type = iter.currentSegment(point);
            iter.next();
            if (type != 4) {
                if (prevPoint != null) {
                    awtLines.add(new java.awt.geom.Line2D.Double(prevPoint[0], prevPoint[1], point[0], point[1]));
                }
            }
            prevPoint = point;
        }

        PathIterator iter2 = shape.getPathIterator(null);
        prevPoint = null;
        double firstPoint[] = null;
        while(!iter2.isDone()) {
            double point[] = new double[2];
            int type = iter2.currentSegment(point);
            iter2.next();
            if (type != 4) {
                if (prevPoint != null) {
                    awtLines.add(new java.awt.geom.Line2D.Double(prevPoint[0], prevPoint[1], point[0], point[1]));
                }
                if (firstPoint == null) {
                    firstPoint = point;
                }
            } else {
                awtLines.add(new java.awt.geom.Line2D.Double(prevPoint[0], prevPoint[1], firstPoint[0], firstPoint[1]));
            }
            prevPoint = point;
        }

        return awtLines;
    }

    protected int calClearDist(int agentX, int agentY, int clearX, int clearY, int length, Set<java.awt.geom.Area> blockades) {
        double REPAIR_COST_FACTOR = 0.000001;
        int counter = 0;
        int min = 0, max = 2 * length;
        while (true) {
            counter++;
            length = (min + max) / 2;
            java.awt.geom.Area area = ExtraActionsUtil.getClearArea(agentX, agentY, clearX, clearY, length,
                    clearRad, 0);

            double firstSurface = Geometry.surface(area);
            for (java.awt.geom.Area blockade : blockades) {
                area.subtract(blockade);
            }
            double surface = Geometry.surface(area);
            double clearedSurface = firstSurface - surface;

            if (clearedSurface * REPAIR_COST_FACTOR > clearRate) {
                max = length;
                continue;
            }

            if (counter != 1 && counter < 15 && max - min > 5) { // 5 is a very
                min = length;
                continue;
            }
            break;
        }
        return length;
    }

    protected List<double[]> getAreaPoints(java.awt.geom.Area area) {
        List<double[]> points = new ArrayList<double[]>();
        PathIterator iter = area.getPathIterator(null);
        while(!iter.isDone()) {
            double point[] = new double[2];
            int type = iter.currentSegment(point);
            if (type == PathIterator.SEG_MOVETO || type == PathIterator.SEG_LINETO) {
                points.add(point);
            }
            iter.next();
        }
        return points;
    }

    protected Blockade isInsideBlockade(int agentX, int agentY, int clearX, int clearY) {
        List<Blockade> blockades = new ArrayList<Blockade>();
        Collection<StandardEntity> entitiesAroundAgent = worldInfo.getObjectsInRange(agentX, agentY, clearDistance);
        for (StandardEntity entity : entitiesAroundAgent) {
            if (entity instanceof Area) {
                Area area = (Area) entity;
                if (area.isBlockadesDefined()) {
                    for (EntityID blockadeID : area.getBlockades()) {
                        Blockade blockade = (Blockade) worldInfo.getEntity(blockadeID);
                        if (blockade == null)
                            continue;
                        if (blockade.getShape() == null) {
                            System.err.println("Blockade Shape is null");
                            continue;
                        }
                        blockades.add(blockade);
                    }
                }
            }
        }
        Map<java.awt.geom.Area, Blockade> blockadeMapping = new HashMap<>();

        for (Blockade blockade : blockades) {
            blockadeMapping.put((new java.awt.geom.Area(blockade.getShape())), blockade);
        }
        int newClearDist = calClearDist(agentX, agentY, clearX, clearY, clearDistance, blockadeMapping.keySet());
        java.awt.geom.Area newClearArea = ExtraActionsUtil.getClearArea(agentX, agentY, clearX, clearY, newClearDist, clearRad, 0);
        List<double[]> newClearAreaPoints = getAreaPoints(newClearArea);
           for (java.awt.geom.Area blockadeArea : blockadeMapping.keySet()) {
            boolean isInside = true;
            for (double[] point : newClearAreaPoints) {
                if (!blockadeArea.contains(point[0], point[1])) {
                    isInside = false;
                    break;
                }
            }
            if (isInside) {
                return blockadeMapping.get(blockadeArea);
            }
        }
        return null;
    }

    protected Action createAction(Point2D currentPos, int clearX, int clearY, Blockade blockade, boolean moveToClear) {
        if (moveToClear) {
            List<EntityID> blockades = new ArrayList<>();
            Collection<EntityID> objectsInClearDist = worldInfo.getObjectIDsInRange(clearX, clearY, clearDistance);
            for (EntityID id : objectsInClearDist) {
                if (worldInfo.getEntity(id) instanceof Road) {
                    Road road = (Road) worldInfo.getEntity(id);
                    if (road != null) {
                        if (road.isBlockadesDefined()) {
                            blockades.addAll(road.getBlockades());
                        }
                    }
                }
            }

            List<Line2D> blockadeLines = new ArrayList<>();
            for (EntityID blockadeID : blockades) {
                Blockade myBlockade = (Blockade)worldInfo.getEntity(blockadeID);
                blockadeLines.addAll(GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(myBlockade.getApexes()), true));
            }

            double minDist = Double.MAX_VALUE;
            Point2D minPoint = null;
            java.awt.geom.Area thinClearArea = ExtraActionsUtil.getClearArea((int)currentPos.getX(), (int)currentPos.getY(),
                    clearX, clearY, clearDistance, passableWidth, 0);

            List<Line2D> clearLines = ExtraActionsUtil.getAreaLines(thinClearArea);

            for (Line2D cLine : clearLines) {
                for (Line2D line : blockadeLines) {
                    Point2D intersection = GeometryTools2D.getSegmentIntersectionPoint(cLine, line);
                    if (intersection != null) {
                        double dist = GeometryTools2D.getDistance(currentPos, intersection);
                        if (dist < minDist) {
                            minDist = dist;
                            minPoint = intersection;
                        }
                    }
                }
            }

            if (minPoint == null) {


                java.awt.geom.Area bigClearArea = ExtraActionsUtil.getClearArea((int)currentPos.getX(), (int)currentPos.getY(),
                        clearX, clearY, clearDistance, clearRad, clearAreaPadding);
                List<Line2D> newClearLines = ExtraActionsUtil.getAreaLines(bigClearArea);

                for (Line2D cLine : newClearLines) {
                    for (Line2D line : blockadeLines) {
                        Point2D intersection = GeometryTools2D.getSegmentIntersectionPoint(cLine, line);
                        if (intersection != null) {
                            double dist = GeometryTools2D.getDistance(currentPos, intersection);
                            if (dist < minDist) {
                                minDist = dist;
                                minPoint = intersection;
                            }
                        }
                    }
                }

                if (minPoint != null) {
                } else {
                    minPoint = new Point2D(clearX, clearY);
                }
            }

                Collection<EntityID> moveEntities = worldInfo.getObjectIDsInRange((int)minPoint.getX(), (int)minPoint.getY(), clearDistance);
                EntityID moveTargetID = null;
                for (EntityID entityID : moveEntities) {
                    if (worldInfo.getEntity(entityID) instanceof Area) {
                        if (ExtraActionsUtil.isInsideArea((Area)worldInfo.getEntity(entityID), (int)minPoint.getX(), (int)minPoint.getY())) {
                            moveTargetID = entityID;
                            break;
                        }
                    }
                }

                if (moveTargetID == null) {

                    minDist = Double.MAX_VALUE;
                    for (EntityID entityID : moveEntities) {
                        if (worldInfo.getEntity(entityID) instanceof Road) {
                            double dist = ExtraActionsUtil.calDistToClosestPointToAreaLines((Area)worldInfo.getEntity(entityID), (int)minPoint.getX(), (int)minPoint.getY());
                            if (dist < minDist) {
                                minDist = dist;
                                moveTargetID = entityID;
                            }
                        }
                    }

                }
                Vector2D toMovePoint = new Vector2D(minPoint.getX()-currentPos.getX(), minPoint.getY()-currentPos.getY());
                int newLength = (int)toMovePoint.getLength()-500; // half meter move away
                toMovePoint = toMovePoint.normalised().scale(newLength);
                minPoint = new Point2D(toMovePoint.getX()+currentPos.getX(), toMovePoint.getY()+currentPos.getY());
                List<EntityID> path = pathPlanning.getResult(agentInfo.getPosition(), moveTargetID);

                if (lastAction instanceof ActionMove) {
                    double tresh = 250; // 25 centimeters
                    double dist = GeometryTools2D.getDistance(lastPosition, new Point2D(agentInfo.getX(), agentInfo.getY()));
                    if (dist < tresh) {
                        Blockade minBlockade = null;
                        minDist = Double.MAX_VALUE;
                        StandardEntity se = worldInfo.getEntity(moveTargetID);
                        if (se instanceof Road && ((Road)se).isBlockadesDefined()) {
                            List<EntityID> moveTargetBlockades = ((Road)se).getBlockades();
                            for (EntityID bID : moveTargetBlockades) {
                                Blockade b = (Blockade)worldInfo.getEntity(bID);
                                if (b != null) {
                                    dist = ExtraActionsUtil.calDistToClosestPointToBlockade(b, (int)agentInfo.getX(), (int)agentInfo.getY());
                                    if (dist < minDist) {
                                        minDist = dist;
                                        minBlockade = b;
                                    }
                                }
                            }
                        }

                        Action newAction = null;
                        if (minBlockade != null) {
                            Point2D agentPoint = new Point2D(agentInfo.getX(), agentInfo.getY());
                            List<Line2D> bLines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(minBlockade.getApexes()));
                            minDist = Double.MAX_VALUE;
                            minPoint = null;
                            for (Line2D line : bLines) {
                                Point2D closestPoint = GeometryTools2D.getClosestPointOnSegment(line, agentPoint);
                                if (closestPoint != null) {
                                    dist = GeometryTools2D.getDistance(closestPoint, agentPoint);
                                    if (dist < minDist && dist < (clearDistance+clearAreaPadding)) {
                                        minDist = dist;
                                        minPoint = closestPoint;
                                    }
                                }
                            }

                            if (minPoint != null) {
                                Point2D longerClear = ExtraActionsUtil.getLongerClearPoint((int)agentInfo.getX(), (int)agentInfo.getY(), (int)minPoint.getX(), (int)minPoint.getY(), clearDistance);
                                newAction = new ActionClear((int)longerClear.getX(), (int)longerClear.getY(), minBlockade);
                            } 
                        }

                        if (newAction != null) {
                            lastAction = newAction;
                        } else {
                            Collection<StandardEntity> entities = worldInfo.getEntitiesOfType(StandardEntityURN.ROAD);
                            StandardEntity randomEntity = Iterables.get(entities, random.nextInt(entities.size()));
                            moveTargetID = randomEntity.getID();
                            List<EntityID> newPath = pathPlanning.getResult(agentInfo.getPosition(),moveTargetID);
                            lastAction = new ActionMove(newPath);
                        }

                        lastPosition = new Point2D(agentInfo.getX(), agentInfo.getY());
                        return lastAction;
                    } 
                }
                if (Double.isNaN(minPoint.getX()) || Double.isNaN(minPoint.getY())) {
                    lastAction = new ActionMove(path);
                } else {
                    lastAction = new ActionMove(path, (int)minPoint.getX(), (int)minPoint.getY());
                }
                lastPosition = new Point2D(agentInfo.getX(), agentInfo.getY());
                return lastAction;
        } else {
            if (lastAction instanceof ActionClear) {
                double stuckTresh = 100;
                int stuckCountTresh = 4;

                Point2D agentPoint = new Point2D(agentInfo.getX(), agentInfo.getY());
                double dist = GeometryTools2D.getDistance(agentPoint, lastPosition);
                if (dist < stuckTresh) {
                    clearStuckCounter++;
                } else {
                    clearStuckCounter = 0;
                }
                if (clearStuckCounter > stuckCountTresh) {
                    List<EntityID> agentPath = new ArrayList<>(1);
                    agentPath.add(agentInfo.getPosition());
                    int maxRandom = 3000;
                    int randomX = random.nextInt(maxRandom)-maxRandom/2;
                    int randomY = random.nextInt(maxRandom)-maxRandom/2;
                    lastAction = new ActionMove(agentPath, (int)agentPoint.getX()+randomX, (int)agentPoint.getY()+randomY);
                    lastPosition = agentPoint;
                    clearStuckCounter = 0;
                    return lastAction;
                }
            }
            Point2D clearPoint = ExtraActionsUtil.getLongerClearPoint((int)agentInfo.getX(), (int)agentInfo.getY(), clearX, clearY, clearDistance);
            lastAction = new ActionClear((int)clearPoint.getX(), (int)clearPoint.getY(), blockade);
            lastPosition = new Point2D(agentInfo.getX(), agentInfo.getY());
            return lastAction;
        }
    }

    protected int clearStuckCounter = 0;

    @Override
    public ExtAction calc() {

        this.result = null;
        PoliceForce policeForce = (PoliceForce)this.agentInfo.me();

        if(this.needRest(policeForce)) {
            List<EntityID> list = new ArrayList<>();
            if(this.target != null) {
                list.add(this.target);
            }
            this.result = this.calcRest(policeForce, this.pathPlanning, list);
            if(this.result != null) {
                return this;
            }
        }

        if(this.target == null) {
            return this;
        }
        EntityID agentPosition = policeForce.getPosition();
        StandardEntity targetEntity = this.worldInfo.getEntity(this.target);
        if(targetEntity == null || !(targetEntity instanceof Area)) {
            return this;
        }

        if (agentInfo.getTime() < scenarioInfo.getKernelAgentsIgnoreuntil())
            return this;

        if (((Human)agentInfo.me()).isBuriednessDefined() && ((Human)agentInfo.me()).getBuriedness() > 0) {
            lastAction = new ActionMove(pathPlanning.getResult(agentInfo.getPosition(), target));
            lastPosition = new Point2D(agentInfo.getX(), agentInfo.getY());

            if (lastAction == null || lastPosition == null) {
                int thresh = 250;
                double dist = GeometryTools2D.getDistance(new Point2D(agentInfo.getX(), agentInfo.getY()), lastPosition);
                if (dist < thresh) {
                    // create random target
                    Collection<StandardEntity> buildings = worldInfo.getEntitiesOfType(StandardEntityURN.BUILDING);
                    StandardEntity randomEntity = Iterables.get(buildings, random.nextInt(buildings.size()));
                    target = randomEntity.getID();
                    lastAction = new ActionMove(pathPlanning.getResult(agentInfo.getPosition(), target));
                }
            }
            result = lastAction;
            return this;
        }

        List<EntityID> pathToTarget = pathPlanning.getResult(agentPosition, target);

        Point2D currentPos = new Point2D(agentInfo.getX(), agentInfo.getY());
        Area currentArea = (Area) worldInfo.getEntity(agentInfo.getPosition());

        //Area nextArea = null;
        Point2D nextWayPoint = null;
        EntityID nextAreaID = pathToTarget.get(0);
        boolean noTarget = false;

        List<Human> humansStuckInBlockades = ExtraActionsUtil.getStuckedHumansSeenByAgent((int) agentInfo.getX(), (int) agentInfo.getY(),
                agentInfo.getID(), worldInfo, agentInfo.getChanged(), scenarioInfo.getPerceptionLosMaxDistance());
        if (humansStuckInBlockades.size() > 0) {
            double minDist = Double.MAX_VALUE;
            Human closestHuman = null;
            for (Human h : humansStuckInBlockades) {
                double dist = GeometryTools2D.getDistance(new Point2D(agentInfo.getX(), agentInfo.getY()), new Point2D(h.getX(), h.getY()));
                if (dist < minDist) {
                    minDist = dist;
                    closestHuman = h;
                }
            }
            if (closestHuman.getPosition().equals(agentInfo.getPosition())) {
                nextWayPoint = new Point2D(closestHuman.getX(), closestHuman.getY());
            } else {
                Edge edge = currentArea.getEdgeTo(nextAreaID);
                if (edge != null) {
                    nextWayPoint = new Point2D((edge.getStartX() + edge.getEndX()) / 2, (edge.getStartY() + edge.getEndY()) / 2);
                } else {
                    nextWayPoint = new Point2D(closestHuman.getX(), closestHuman.getY());
                }
            }
        } else {

            Edge edge = currentArea.getEdgeTo(nextAreaID);

            if (nextAreaID.equals(currentArea.getID())) {
                List<EntityID> blockadeIDs = currentArea.getBlockades();
                double maxSurface = Double.MIN_VALUE;
                EntityID maxBlockadeID = null;
                for (EntityID blockadeID : blockadeIDs) {
                    Blockade b = (Blockade) worldInfo.getEntity(blockadeID);
                    double surface = Geometry.surface(new java.awt.geom.Area(b.getShape()));
                    if (surface > maxSurface) {
                        maxSurface = surface;
                        maxBlockadeID = b.getID();
                    }
                }
                if (maxBlockadeID == null) {
                    noTarget = true;
                } else {
                    result = new ActionClear(maxBlockadeID);
                    return this;
                }
            }

            if (edge == null) {
                nextWayPoint = new Point2D(currentArea.getX(), currentArea.getY());
                noTarget = true;
//            }
            } else {
                nextWayPoint = new Point2D((edge.getStartX() + edge.getEndX()) / 2, (edge.getStartY() + edge.getEndY()) / 2);
            }
        }

        boolean moveToClear = false;

        List<java.awt.geom.Line2D> debugClearAreaLines = new ArrayList<>();
        int loopCounter = 0;

        while (!noTarget) {
            Vector2D clearVector = ExtraActionsUtil.getClearVector((int)currentPos.getX(), (int)currentPos.getY(),
                    (int)nextWayPoint.getX(), (int)nextWayPoint.getY(), clearDistance, clearAreaPadding);
            Blockade encapsulatingBlockade = isInsideBlockade((int) currentPos.getX(), (int) currentPos.getY(), (int)clearVector.getX(), (int)clearVector.getY());
            if (encapsulatingBlockade != null) {
                result = new ActionClear(encapsulatingBlockade.getID());
                return this;
            } else {
                if (MIC_ATConfig.isStuckInBlockade((int)agentInfo.getX(), (int)agentInfo.getY(), agentInfo.getPosition(),worldInfo)) {
                    Collection<Blockade> blockades = worldInfo.getBlockades(agentInfo.getPosition());
                    Point2D agentPoint = ExtraActionsUtil.getAgentPosition(agentInfo);
                    Blockade minBlockade = null;
                    double minDist = Double.MAX_VALUE;
                    for (Blockade b : blockades) {
                        List<Line2D> bLines = ExtraActionsUtil.getAllBlockadeLines(b, worldInfo);
                        for (Line2D line : bLines) {
                            Point2D interPoint = GeometryTools2D.getClosestPointOnSegment(line, agentPoint);
                            if (interPoint != null) {
                                double dist = GeometryTools2D.getDistance(interPoint, agentPoint);
                                if (dist < minDist) {
                                    minDist = dist;
                                    minBlockade = b;
                                }
                            }
                        }
                    }
                    result = new ActionClear(minBlockade.getID());
                    return this;
                }
            }

            Collection<EntityID> entitiesInRange = worldInfo.getObjectIDsInRange((int)currentPos.getX(),(int)currentPos.getY(), clearDistance);
            List<EntityID> blockades = new ArrayList<EntityID>();
            for (EntityID entityID : entitiesInRange) {
                if (worldInfo.getEntity(entityID) instanceof Road) {
                    List<EntityID> newBlockades = ((Road) worldInfo.getEntity(entityID)).getBlockades();
                    if (newBlockades != null) {
                        blockades.addAll(newBlockades);
                    }
                }
            }

            java.awt.geom.Area clearArea = ExtraActionsUtil.getClearArea((int)currentPos.getX(), (int)currentPos.getY(),
                    (int)nextWayPoint.getX(), (int)nextWayPoint.getY(), clearDistance, passableWidth, clearAreaPadding);

            List<Line2D> clearLines = ExtraActionsUtil.getAreaLines(clearArea);
            debugClearAreaLines.addAll(convertLinesToAwtLines(clearLines));

            for (EntityID blockadeID : blockades) {
                Blockade blockade = (Blockade) worldInfo.getEntity(blockadeID);
                List<Line2D> blockadeLines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(blockade.getApexes()), true);
                for (Line2D blockadeLine : blockadeLines) {
                    for (Line2D clearLine : clearLines) {
                        Point2D intersection = GeometryTools2D.getSegmentIntersectionPoint(clearLine, blockadeLine);
                        if (intersection != null) {
                   

                            result = createAction(currentPos, (int)clearVector.getX(), (int)clearVector.getY(), blockade, moveToClear);
                            return this;
                        }
                    }
                }

                for (Line2D blockadeLine : blockadeLines) {
                    if (clearArea.contains(blockadeLine.getOrigin().getX(), blockadeLine.getOrigin().getY())) {
                        result = createAction(currentPos, (int)clearVector.getX(), (int)clearVector.getY(), blockade, moveToClear);
                        return this;
                    }
                }

            }
            EntityID newTarget = null;
            Point2D minIntersectionPoint = null;

            if (ExtraActionsUtil.isInsideArea(currentArea, (int)clearVector.getX(), (int)clearVector.getY())) {
                newTarget = currentArea.getID();
            } else {
                double minDist = Double.MAX_VALUE;
                Line2D newClearLine = new Line2D(new Point2D(currentPos.getX(), currentPos.getY()), new Point2D(clearVector.getX(), clearVector.getY()));
                int i = pathToTarget.indexOf(currentArea.getID());
                if (i < 0) {
                    i = 0;
                }
                for (; i < pathToTarget.size(); i++) {
                    StandardEntity se = worldInfo.getEntity(pathToTarget.get(i));
                    if (se instanceof Area && !se.getID().equals(currentArea.getID())) {
                        Area area = (Area) se;
                        if (area.isXDefined() && area.isYDefined()) {
                            List<Line2D> roadLines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(area.getApexList()), true);
                            for (Line2D line : roadLines) {
                                Point2D intersection = GeometryTools2D.getSegmentIntersectionPoint(line, newClearLine);
                                if (intersection != null) {
                                    double dist = GeometryTools2D.getDistance(intersection, currentPos);
                                    if (dist < minDist) {
                                        minDist = dist;
                                        newTarget = se.getID();
                                        minIntersectionPoint = intersection;
                                    }
                                }
                            }
                        }
                    }
                }
                if (newTarget == null) {
                }
            }

            if (newTarget == null) {
                    result = createAction(currentPos, (int)nextWayPoint.getX(), (int)nextWayPoint.getY(), null, true);
                    return this;
            } else {
                moveToClear = true;

                if (newTarget.equals(currentArea.getID())) {
                    double thresh = 100;
                    double dist = GeometryTools2D.getDistance(currentPos, nextWayPoint);
                    if (dist < thresh) {
                        break;
                    } else {
                        Point2D clearPoint = new Point2D(clearVector.getX(), clearVector.getY());

                        double clearDist = GeometryTools2D.getDistance(clearPoint, currentPos);
                        double nextWayPointDist = GeometryTools2D.getDistance(nextWayPoint, currentPos);

                        if (clearDist < nextWayPointDist) {
                            currentPos = new Point2D(clearVector.getX(), clearVector.getY());
                        } else {
                            if (currentArea.getID().equals(pathToTarget.get(pathToTarget.size()-1))) {
                                break;
                            }
                        }
                    }
                } else {

                    int i = 0;
                    for (; i < pathToTarget.size(); i++) {
                        if (pathToTarget.get(i).equals(newTarget)) {
                            break;
                        }
                    }
                    if (i + 1 >= pathToTarget.size()) {
                        i = pathToTarget.size() - 1;
                    } else {
                        i++;
                    }

                    EntityID newNextAreaID = pathToTarget.get(i);
                    currentArea = (Area) worldInfo.getEntity(newTarget);
                    Edge edge = currentArea.getEdgeTo(newNextAreaID);

                    if (edge != null) {
                        currentPos = minIntersectionPoint;
                        nextWayPoint = new Point2D((edge.getStartX() + edge.getEndX()) / 2, (edge.getStartY() + edge.getEndY()) / 2);
                    } else {
                        if (newNextAreaID.equals(newTarget)) {
                            nextWayPoint = new Point2D(currentArea.getX(), currentArea.getY());
                        } 
                    }
                }
            }
            loopCounter++;
            if (loopCounter == 500) {
                break;
            }
        }

        this.result = new ActionMove(pathToTarget);
        return this;
    }

    protected String printPath(List<EntityID> path) {
        StringBuilder sb = new StringBuilder();
        sb.append("[");
        for (EntityID id : path) {
            sb.append(id+ ",");
        }
        sb.append("]");
        return sb.toString();
    }

    private boolean needRest(Human agent) {
        int hp = agent.getHP();
        int damage = agent.getDamage();
        if(damage == 0 || hp == 0) {
            return false;
        }
        int activeTime = (hp / damage) + ((hp % damage) != 0 ? 1 : 0);
        if(this.kernelTime == -1) {
            try {
                this.kernelTime = this.scenarioInfo.getKernelTimesteps();
            }catch (NoSuchConfigOptionException e) {
                this.kernelTime = -1;
            }
        }
        return damage >= this.thresholdRest || (activeTime + this.agentInfo.getTime()) < this.kernelTime;
    }

    private Action calcRest(Human human, PathPlanning pathPlanning, Collection<EntityID> targets) {
        EntityID position = human.getPosition();
        Collection<EntityID> refuges = this.worldInfo.getEntityIDsOfType(StandardEntityURN.REFUGE);
        int currentSize = refuges.size();
        if(refuges.contains(position)) {
            //debugUtil.debug("refuges contains position:" + position);
            return new ActionRest();
        }
        List<EntityID> firstResult = null;
        while(refuges.size() > 0) {
            pathPlanning.setFrom(position);
            pathPlanning.setDestination(refuges);
            List<EntityID> path = pathPlanning.calc().getResult();
            if (path != null && path.size() > 0) {
                if (firstResult == null) {
                    firstResult = new ArrayList<>(path);
                    if(targets == null || targets.isEmpty()) {
                        break;
                    }
                }
                EntityID refugeID = path.get(path.size() - 1);
                pathPlanning.setFrom(refugeID);
                pathPlanning.setDestination(targets);
                List<EntityID> fromRefugeToTarget = pathPlanning.calc().getResult();
                if (fromRefugeToTarget != null && fromRefugeToTarget.size() > 0) {
                    //debugUtil.debug("calculated a path to refugee to rest, refugee:" + position);
                    return new ActionMove(path);
                }
                refuges.remove(refugeID);
                //remove failed
                if (currentSize == refuges.size()) {
                    break;
                }
                currentSize = refuges.size();
            } else {
                break;
            }
        }
        return firstResult != null ? new ActionMove(firstResult) : null;
    }
}
