package MIC.module.complex;



import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.StandardMessagePriority;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.complex.BuildingDetector;
import org.apache.commons.math3.random.MersenneTwister;
import org.apache.commons.math3.random.RandomGenerator;

import MIC.module.algorithm.MICKMeans;
import MIC.module.algorithm.MICPathPlanning;
import MIC.util.DistanceSorter;
import MIC.util.Agent.MIC_FBConfig;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;
import java.util.Map.Entry;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class MICBuildingDetector extends BuildingDetector {
    private EntityID result;

    private MICKMeans clustering;
    private MIC_FBConfig fireConfig;

    private int sendTime;
    private int sendingAvoidTimeClearRequest;

    private Collection<EntityID> agentPositions;
    private Map<EntityID, Integer> sentTimeMap;
    private int sendingAvoidTimeReceived;
    private int sendingAvoidTimeSent;

    private int moveDistance;
    private EntityID lastPosition;
    private int positionCount;

    private RandomGenerator random;
    private MICPathPlanning pathPlanning;

    protected ArrayList<StandardEntity> all_result = new ArrayList<>();
    protected List<StandardEntity> fireAgents;
    protected List<StandardEntity> refuges;
    protected StandardEntity RefugeAgentID;
    protected List<StandardEntity> visitedRefugeBuildings = new ArrayList<StandardEntity>();
    protected int amghezi; 
    protected int lucas;
    protected Boolean refugeMode = true;
    protected StandardEntity refugeBuilding;

    public MICBuildingDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        switch  (si.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.clustering = moduleManager.getModule("MICBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
            case PRECOMPUTED:
                this.clustering = moduleManager.getModule("MICBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
            case NON_PRECOMPUTE:
                this.clustering = moduleManager.getModule("MICBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
        }
        this.sendTime = 0;
        this.sendingAvoidTimeClearRequest = developData.getInteger("MICBuildingDetector.sendingAvoidTimeClearRequest", 5);

        this.agentPositions = new HashSet<>();
        this.sentTimeMap = new HashMap<>();
        this.sendingAvoidTimeReceived = developData.getInteger("MICBuildingDetector.sendingAvoidTimeReceived", 3);
        this.sendingAvoidTimeSent = developData.getInteger("MICBuildingDetector.sendingAvoidTimeSent", 5);

        this.moveDistance = developData.getInteger("MICBuildingDetector.moveDistance", 40000);
        int randomSeed = developData.getInteger("MICBuildingDetector.randomSeed", 13);
        random = new MersenneTwister(randomSeed);
        
        this.amghezi = 0;
        this.lucas = 0;
        fireConfig = new MIC_FBConfig(this.worldInfo);
        
        all_result = new ArrayList<>();
        fireAgents = new ArrayList<>(worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE));
        refuges = new ArrayList<StandardEntity>(this.worldInfo.getEntitiesOfType(REFUGE));
        this.RefugeAgentID = fireAgents.get(0);

    }

    @Override
    public BuildingDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.clustering.updateInfo(messageManager);

        this.reflectMessage(messageManager);
        this.sendEntityInfo(messageManager);

        if(this.result != null) {
            Building building = (Building)this.worldInfo.getEntity(this.result);
            if(building.isFierynessDefined())
            	if(building.getFieryness() >= 4) {
                messageManager.addMessage(new MessageBuilding(true, building));
            }
        }
        
        int currentTime = this.agentInfo.getTime();
        Human agent = (Human)this.agentInfo.me();
        int agentX = agent.getX();
        int agentY = agent.getY();
        StandardEntity positionEntity = this.worldInfo.getPosition(agent);
        if(positionEntity instanceof Road) {
            Road road = (Road)positionEntity;
            if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
                for(Blockade blockade : this.worldInfo.getBlockades(road)) {
                    if(blockade == null || !blockade.isApexesDefined()) {
                        continue;
                    }
                    if(fireConfig.isInsideApex(agentX, agentY, blockade.getApexes())) {
                            this.sendTime = currentTime;
                            messageManager.addMessage(
                                    new CommandPolice(
                                            true,
                                            StandardMessagePriority.HIGH,
                                            null,
                                            agent.getPosition(),
                                            CommandPolice.ACTION_CLEAR
                                    )
                            );

                            break;
                    }
                }
            }
            if(this.lastPosition != null && this.lastPosition.getValue() == road.getID().getValue()) {
                this.positionCount++;
                if(this.positionCount > this.getMaxTravelTime(road)) {
                    if ((this.sendTime + this.sendingAvoidTimeClearRequest) <= currentTime) {
                        this.sendTime = currentTime;
                    }
                }
            } else {
                this.lastPosition = road.getID();
                this.positionCount = 0;
            }
        }
        return this;
    }

    protected Point2D getMaxPoint(Road road) {
        List<Point2D> points = GeometryTools2D.vertexArrayToPoints(road.getApexList());
        int maxX = Integer.MIN_VALUE;
        int maxY = Integer.MIN_VALUE;
        for (Point2D point : points) {
            if (point.getX() > maxX) {
                maxX = (int)point.getX();
            }
            if (point.getY() > maxY) {
                maxY = (int)point.getY();
            }
        }
        return new Point2D(maxX, maxY);
    }


    protected Point2D getMinPoint(Road road) {
        List<Point2D> points = GeometryTools2D.vertexArrayToPoints(road.getApexList());
        int minX = Integer.MAX_VALUE;
        int minY = Integer.MAX_VALUE;
        for (Point2D point : points) {
            if (point.getX() < minX) {
                minX = (int)point.getX();
            }
            if (point.getY() < minY) {
                minY = (int)point.getY();
            }
        }
        return new Point2D(minX, minY);
    }

    

    public List<StandardEntity> detectRefugeBuildings(StandardEntity refuge)
    {
    	
    	List<StandardEntity> FPBuildings = new ArrayList<StandardEntity>();
    	List<EntityID> buildings = selectNearbyBuildings (refuge.getID());
		Map<StandardEntity, Double> BuildingDistances = new HashMap<StandardEntity, Double>();
			
		for(EntityID building : buildings)
		{
			
			StandardEntity standardBuilding = this.worldInfo.getEntity(building);
    		
			if(building != null)
			{   
				double distance = this.worldInfo.getDistance(building, refuge.getID());
				BuildingDistances.put(standardBuilding, distance);
    		}
		}
		
		Map<StandardEntity, Double> sorted = this.CurrentSortMap(BuildingDistances, true);
		List<StandardEntity> finalList = new ArrayList<StandardEntity>(sorted.keySet());
		
		for(Iterator<StandardEntity> iterator = finalList.iterator(); iterator.hasNext();)
		{        		
			StandardEntity object = iterator.next();
			if(object != finalList.get(0)
					&& object != finalList.get(1)
					&& object != finalList.get(2)
					&& object != finalList.get(3))
				iterator.remove();
		}
		
		FPBuildings.addAll(finalList);		
		return FPBuildings;
    }
    
    public Boolean needBreak(List<EntityID> neighbours)
    {
    	int count = 0;
    	for(EntityID se : neighbours)
    	{
    		StandardEntity e = this.worldInfo.getEntity(se);
    		if(e instanceof Building)
    			count += 1;
    	}
    	if(count == 0) return true;
    	else return false;
    }
    
    public List<EntityID> removeListFromList(List<EntityID> source, List<EntityID> deleted)
    {
    	for( int i = 0 ; i < deleted.size() ; i ++)
    	{
        	for(Iterator<EntityID> itr = source.iterator(); itr.hasNext();)	
        	{
        		if( deleted.get(i) == itr.next())
        			itr.remove();
        	}
    	}
    	return source;
    }
    
    
    List<EntityID> repeatedNeighbours = new ArrayList<EntityID>();
    
    public List<StandardEntity> detectRefugeNeighbour(StandardEntity refuge)
    { 
    	if(refuge instanceof Refuge)
    		repeatedNeighbours.add(refuge.getID());    	
    	List<StandardEntity> result = new ArrayList<StandardEntity>();
    	Building b = (Building) refuge;
    	List<EntityID> neighbours = b.getNeighbours();
    	if(!neighbours.isEmpty())
	    	for(EntityID eid : neighbours)
	    	{
	    		if(this.worldInfo.getEntity(eid) instanceof Building)
	    		{
	    			ArrayList<EntityID> ce = (ArrayList<EntityID>) (((Building) this.worldInfo.getEntity(eid)).getNeighbours());
	    			removeListFromList(ce, repeatedNeighbours);
	    			if(needBreak(ce))
	    			{
	    				result.add(this.worldInfo.getEntity(eid));
	    				continue;
	    			}
	    			repeatedNeighbours.add(eid);
	    			result.add(this.worldInfo.getEntity(eid));
	    			result.addAll(detectRefugeNeighbour( this.worldInfo.getEntity(eid) ));
	    			
	    		}
	    	}
    	
    	return result;
    	
    } 

	public List<EntityID> selectNearbyBuildings(EntityID Refuge)
	{
		List<EntityID> nearby = new ArrayList<EntityID>();
		List<EntityID> nearbyBuildings = new ArrayList<EntityID>();
		
		int range = 200;
	
		while(nearbyBuildings.size() < 20)
		{
			
			nearby = new ArrayList<EntityID>(this.worldInfo.getObjectIDsInRange(Refuge, range));
		
			for(EntityID ei : nearby)
			{
				StandardEntity se = this.worldInfo.getEntity(ei);
				if ( nearbyBuildings.contains(ei) )
					continue;
				
				List<StandardEntity> buildings = new ArrayList<StandardEntity>(this.worldInfo.getEntitiesOfType(BUILDING));
				if(buildings.contains( se ))
				{
					nearbyBuildings.add( se.getID() );
				}
			}
			if(nearbyBuildings.size() < 20)
				range += 100;	
		}
		
		return nearbyBuildings;
	}
    
    

    public Map<StandardEntity, Double> CurrentSortMap(Map<StandardEntity, Double> unsortedMap, boolean order)
	{
		List<Entry<StandardEntity, Double>> list = new LinkedList<Entry<StandardEntity, Double>>(unsortedMap.entrySet());
		Collections.sort(list, new Comparator<Entry<StandardEntity, Double>>()
		{
			
			public int compare (Entry<StandardEntity, Double> o1,
					Entry<StandardEntity, Double> o2)
			{
					if(order)
						return o1.getValue().compareTo(o2.getValue());
					else
						return o2.getValue().compareTo(o1.getValue());
			}
			
		}
		);
		
		Map<StandardEntity, Double> sortedMap = new LinkedHashMap<StandardEntity, Double>();
		for( Entry<StandardEntity, Double> entry : list )
		{
			sortedMap.put(entry.getKey(), entry.getValue());
		}
		
		return sortedMap;
	}

    public List<StandardEntity> removeFromList(List<StandardEntity> list, StandardEntity deletable)
    {
    	Iterator<StandardEntity> itr = list.iterator();
    	
    	while(itr.hasNext())
    	{
			if(itr.next() == deletable)
				itr.remove();
    	}
       	return list;
    }
    
    
    @Override
    public BuildingDetector calc() {
    	if(this.refugeMode)
    	{
    		if(this.fireAgents.size() >= 5)
        	{
        		if(this.agentInfo.getID().getValue() == this.RefugeAgentID.getID().getValue())
            	{            		
        		    StandardEntity myResult = null;
        	    	if(all_result.isEmpty())
        			{
        	    		
        	    		this.amghezi += 1;
        	    		
        	    		for(int i = 0 ; i < refuges.size() ; i++)
        	    		{
        	    			all_result = new ArrayList<StandardEntity>();
        	    			all_result.addAll(this.detectRefugeBuildings(refuges.get(i)));        	    			
        	    			
        	    			if(all_result.isEmpty() == false)
            	    			break;
        	    		}
        	    		
        	    		if(all_result.isEmpty())
        	    		{
        	    			this.refugeMode = false;
        	    		}
        	    		
        	    		all_result.sort(new DistanceSorter(this.worldInfo, refuges.get(0)));
        	    		StandardEntity repeatEntity = refuges.get(0);
        				removeFromList(refuges, refuges.get(0));
        				refuges.add(repeatEntity);
        				if(!all_result.isEmpty())
        					myResult = all_result.get(0);
        				        				
        				this.refugeBuilding = myResult;
        				if(!all_result.isEmpty())
        					removeFromList(all_result, all_result.get(0));
        				
        				
        				
        				if(this.amghezi >= this.refuges.size())
        					this.refugeMode = false;
        	    		
        			}
        	    	else
        	    	{
        	    		int n = 0;
        	    		Building bb = (Building) refugeBuilding;
        	    		
        	    		if(bb != null)
        	    		{
        	    			StandardEntityURN positionURN = this.worldInfo.getEntity(this.agentInfo.getPosition()).getStandardURN();
        	    			if(positionURN == StandardEntityURN.REFUGE)
        	    				lucas += 1;
        	    			
        	    			if(this.lucas > this.agentInfo.getWater())
            	    		{
            	    			myResult = all_result.get(0);
            	    			removeFromList(all_result, all_result.get(0));
            	    			this.lucas = this.agentInfo.getWater();

            	    			if(visitedRefugeBuildings.size() > 0)
            	    			{
            	    				while(visitedRefugeBuildings.contains(myResult))
                	    			{
                	    				if( n < all_result.size() - 1 )
                	    				{		
                	    					n += 1;
                	    					myResult = all_result.get(n);
                	    				}
                	    				else
                	    					all_result.clear();
                	    			}	
            	    			}
            	    		}
            	    		else
            	    			myResult = all_result.get(0);
        	    		}
        	    		
        	    		
        	    	}

        	    	if(myResult != null)
        			{
        	    		this.lucas = this.agentInfo.getWater();
        				this.result = myResult.getID();
        				this.visitedRefugeBuildings.add(myResult);
            			return this;
        			}
        			else
        			{        				
        				this.result = this.refugeBuilding.getID();
            			return this;
        			}
    			
            	}
        		
        	}
    	}
    	
    	this.result = fireConfig.firstStage(this.worldInfo, fireAgents, this.agentInfo);
    	if(this.result == null)
    		this.result = fireConfig.secondStage(this.worldInfo, this.agentInfo);
    	
        return this;
    }

    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public BuildingDetector precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        return this;
    }

    @Override
    public BuildingDetector resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        preparate();
        return this;
    }

    @Override
    public BuildingDetector preparate() {
        super.preparate();
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.clustering.preparate();
        return this;
    }

    private void reflectMessage(MessageManager messageManager) {
        Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
        changedEntities.add(this.agentInfo.getID());
        int time = this.agentInfo.getTime();
        for(CommunicationMessage message : messageManager.getReceivedMessageList(MessageBuilding.class)) {
            MessageBuilding mb = (MessageBuilding)message;
            if(!changedEntities.contains(mb.getBuildingID())) {
                MessageUtil.reflectMessage(this.worldInfo, mb);
            }
            this.sentTimeMap.put(mb.getBuildingID(), time + this.sendingAvoidTimeReceived);
        }
    }

    private boolean checkSendFlags(){
        boolean isSendBuildingMessage = true;

        StandardEntity me = this.agentInfo.me();
        if(!(me instanceof Human)){
            return false;
        }
        Human agent = (Human)me;
        EntityID agentID = agent.getID();
        EntityID position = agent.getPosition();
        StandardEntityURN agentURN = agent.getStandardURN();
        EnumSet<StandardEntityURN> agentTypes = EnumSet.of(AMBULANCE_TEAM, FIRE_BRIGADE, POLICE_FORCE);
        agentTypes.remove(agentURN);

        this.agentPositions.clear();
        for(StandardEntity entity : this.worldInfo.getEntitiesOfType(agentURN)) {
            Human other = (Human)entity;
            if(isSendBuildingMessage) {
                if (other.getPosition().getValue() == position.getValue()) {
                    if (other.getID().getValue() > agentID.getValue()) {
                        isSendBuildingMessage = false;
                    }
                }
            }
            this.agentPositions.add(other.getPosition());
        }
        for(StandardEntityURN urn : agentTypes) {
            for(StandardEntity entity : this.worldInfo.getEntitiesOfType(urn)) {
                Human other = (Human) entity;
                if(isSendBuildingMessage) {
                    if (other.getPosition().getValue() == position.getValue()) {
                        if (urn == FIRE_BRIGADE) {
                            isSendBuildingMessage = false;
                        } else if (agentURN != FIRE_BRIGADE && other.getID().getValue() > agentID.getValue()) {
                            isSendBuildingMessage = false;
                        }
                    }
                }
                this.agentPositions.add(other.getPosition());
            }
        }
        return isSendBuildingMessage;
    }

    private void sendEntityInfo(MessageManager messageManager) {
        if(this.checkSendFlags()) {
            Building building = null;
            int currentTime = this.agentInfo.getTime();
            Human agent = (Human) this.agentInfo.me();
            for (EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
                StandardEntity entity = this.worldInfo.getEntity(id);
                if (entity instanceof Building) {
                    Integer time = this.sentTimeMap.get(id);
                    if (time != null && time > currentTime) {
                        continue;
                    }
                    Building target = (Building) entity;
                    if (!this.agentPositions.contains(target.getID())) {
                        building = this.selectBuilding(building, target);
                    } else if (target.getID().getValue() == agent.getPosition().getValue()) {
                        building = this.selectBuilding(building, target);
                    }
                }
            }
            if (building != null) {
                messageManager.addMessage(new MessageBuilding(true, building));
                this.sentTimeMap.put(building.getID(), currentTime + this.sendingAvoidTimeSent);
            }
        }
    }

    private Building selectBuilding(Building building1, Building building2) {
        if(building1 != null) {
            if(building2 != null) {
                if(building1.isOnFire() && building2.isOnFire()) {
                    if (building1.getFieryness() < building2.getFieryness()) {
                        return building2;
                    } else if (building1.getFieryness() > building2.getFieryness()) {
                        return building1;
                    }
                    if(building1.isTemperatureDefined() && building2.isTemperatureDefined()) {
                        return building1.getTemperature() < building2.getTemperature() ? building2 : building1;
                    }
                } else if (!building1.isOnFire() && building2.isOnFire()) {
                    return building2;
                }
            }
            return building1;
        }
        return building2 != null ? building2 : null;
    }

    private int getMaxTravelTime(Area area) {
        int distance = 0;
        List<Edge> edges = new ArrayList<>();
        for(Edge edge : area.getEdges()) {
            if(edge.isPassable()) {
                edges.add(edge);
            }
        }
        if(edges.size() <= 1) {
            return Integer.MAX_VALUE;
        }
        for(int i = 0; i < edges.size(); i++) {
            for(int j = 0; j < edges.size(); j++) {
                if(i != j) {
                    Edge edge1 = edges.get(i);
                    double midX1 = (edge1.getStartX() + edge1.getEndX()) / 2;
                    double midY1 = (edge1.getStartY() + edge1.getEndY()) / 2;
                    Edge edge2 = edges.get(j);
                    double midX2 = (edge2.getStartX() + edge2.getEndX()) / 2;
                    double midY2 = (edge2.getStartY() + edge2.getEndY()) / 2;
                    int d = this.getDistance(midX1, midY1, midX2, midY2);
                    if(distance < d) {
                        distance = d;
                    }
                }
            }
        }
        if(distance > 0) {
            return (distance / this.moveDistance) + ((distance % this.moveDistance) > 0 ? 1 : 0) + 1;
        }
        return Integer.MAX_VALUE;
    }

    private int getDistance(double fromX, double fromY, double toX, double toY) {
        double dx = toX - fromX;
        double dy = toY - fromY;
        return (int)Math.hypot(dx, dy);
    }
}