package MIC.module.complex;

import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.information.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import adf.debug.TestLogger;
import org.apache.log4j.Logger;

import adf.component.communication.CommunicationMessage;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.complex.Search;

import rescuecore2.standard.entities.*;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

import MIC.util.*;


public class MICFireSearch extends Search {
	private PathPlanning pathPlanning;
	private Clustering clustering;
	private Set<EntityID> emptyBuildings;
	private Set<EntityID> burnBuildings;
	private EntityID result;
	private Collection<EntityID> unsearchedBuildingIDs;
	private List<EntityID> unsearchedRoadIDs;
	private Collection<EntityID> blockedBuildingIDs;
	
	private CollectionUtil collectionUtil;
	
	private int threshold=5;
	private int SearchTime=threshold;
	private int SearchclusterIndex=0;
	
	private boolean DangerExists = false;
	private boolean Distributed=false;
	private boolean BurnOut=false;
	
	public MICFireSearch(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
		super(ai, wi, si, moduleManager, developData);
		this.unsearchedBuildingIDs = new HashSet<>();
		this.unsearchedRoadIDs=new ArrayList<>();
		this.emptyBuildings=new HashSet<>();
		this.burnBuildings=new HashSet<>();
		this.blockedBuildingIDs=new HashSet<>();
		StandardEntityURN agentURN = ai.me().getStandardURN();
		if (agentURN == StandardEntityURN.AMBULANCE_TEAM) {
			this.pathPlanning = moduleManager.getModule("MICSearch.PathPlanning.Ambulance", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("MICSearch.Clustering.Ambulance", "adf.sample.module.algorithm.SampleKMeans");
		} else if (agentURN == StandardEntityURN.FIRE_BRIGADE) {
			this.pathPlanning = moduleManager.getModule("MICSearch.PathPlanning.Fire", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("MICSearch.Clustering.Fire", "adf.sample.module.algorithm.SampleKMeans");
		} else if (agentURN == StandardEntityURN.POLICE_FORCE) {
			this.pathPlanning = moduleManager.getModule("MICSearch.PathPlanning.Police", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("MICSearch.Clustering.Police", "adf.sample.module.algorithm.SampleKMeans");
		}
		
		registerModule(this.clustering);
		registerModule(this.pathPlanning);
		this.collectionUtil = new CollectionUtil();
	}

	@Override
	public Search updateInfo(MessageManager messageManager) {

		super.updateInfo(messageManager);
		this.reflectMessage(messageManager);

		this.unsearchedRoadIDs.removeAll(this.worldInfo.getChanged().getChangedEntities());
		
		
		this.DangerExists = this.firstStage();
		this.BurnOut = this.secondStage();
		
		if(this.DangerExists)
			this.SearchTime=threshold;
		
		if(BurnOut)
		{
			this.SearchTime=0;
			reset();
		}
		
		if(this.unsearchedRoadIDs.isEmpty())
            this.reset();
		
        this.unsearchedRoadIDs.removeAll(this.worldInfo.getChanged().getChangedEntities());

        if(this.unsearchedRoadIDs.isEmpty())
        {
            this.reset();
            this.unsearchedRoadIDs = this.collectionUtil.removeIDListFromList(
            		this.unsearchedRoadIDs,
            		new ArrayList<EntityID>(this.worldInfo.getChanged().getChangedEntities()) );
        }
        return this;
	}

	@Override
	public Search calc() {
		if(this.agentInfo.me() instanceof FireBrigade)
		{
			this.result = null;
			if (unsearchedRoadIDs.isEmpty())
				return this;
			this.pathPlanning.setFrom(this.agentInfo.getPosition());
			this.pathPlanning.setDestination(this.unsearchedRoadIDs);
			List<EntityID> path = this.pathPlanning.calc().getResult();
			if (path != null) {
				this.result = path.get(path.size() - 1);
			}
			return this;
		}
		else
		{
			this.result = null;
			if (unsearchedBuildingIDs.isEmpty())
				return this;

			this.pathPlanning.setFrom(this.agentInfo.getPosition());
			this.pathPlanning.setDestination(this.unsearchedBuildingIDs);
			List<EntityID> path = this.pathPlanning.calc().getResult();
			if (path != null && path.size() > 2) {
				this.result = path.get(path.size() - 2);
			}else if (path != null && path.size() > 0) {
				this.result = path.get(path.size() - 1);
			}
			return this;
		}
	}
	
	private void reset() {
		this.firstStage();
		if(this.Distributed==false)
		{
			this.SearchclusterIndex=this.clustering.getClusterIndex(this.agentInfo.getID());
			this.Distributed=true;
		}
		if(this.agentInfo.me() instanceof FireBrigade)
		{
			if(this.DangerExists == true)
				this.SearchTime = threshold;
			else
				this.SearchTime -= 1;
			
			if(this.SearchTime > 0)
			{
				this.unsearchedRoadIDs.clear();
		        Collection<StandardEntity> clusterEntities = null;
		        if(this.clustering != null) {
		            clusterEntities = this.clustering.getClusterEntities(this.SearchclusterIndex);
		        }
		        if(clusterEntities != null && clusterEntities.size() > 0) 
		        {
		            for(StandardEntity entity : clusterEntities) 
		            {
		            	if(entity instanceof Road)
		            	{	Road road=(Road) entity;
		            		boolean ifadd=true;
		            		for(EntityID entityid:road.getNeighbours())
		            		{
		            			if(this.worldInfo.getEntity(entityid) instanceof Building)
		            			{	
		            				ifadd=false;
		            				break;
		            			}
		            		}
		            		if(ifadd)
		            			this.unsearchedRoadIDs.add(entity.getID());
		            	}
		            }
		        } 
		        else 
		        {
		            this.unsearchedRoadIDs.addAll(this.worldInfo.getEntityIDsOfType(
		                    StandardEntityURN.ROAD,
		                    StandardEntityURN.GAS_STATION,
		                    StandardEntityURN.AMBULANCE_CENTRE,
		                    StandardEntityURN.FIRE_STATION,
		                    StandardEntityURN.POLICE_OFFICE
		            ));
		        }
			}
			else
			{
				this.SearchTime=threshold;
				this.SearchclusterIndex=(this.SearchclusterIndex+1)%this.clustering.getClusterNumber();
				this.unsearchedRoadIDs.clear();
		        Collection<StandardEntity> clusterEntities = null;
		        if(this.clustering != null) {
		            clusterEntities = this.clustering.getClusterEntities(this.SearchclusterIndex);
		        }
		        if(clusterEntities != null && clusterEntities.size() > 0) 
		        {
		            for(StandardEntity entity : clusterEntities) 
		            {
		            	if(entity instanceof Road)
		            	{this.unsearchedRoadIDs.add(entity.getID());}
		            }
		        } 
		        else 
		        {
		            this.unsearchedRoadIDs.addAll(this.worldInfo.getEntityIDsOfType(
		                    StandardEntityURN.ROAD,
		                    StandardEntityURN.GAS_STATION,
		                    StandardEntityURN.AMBULANCE_CENTRE,
		                    StandardEntityURN.FIRE_STATION,
		                    StandardEntityURN.POLICE_OFFICE
		            ));
		        }
			}
		}
		else
		{
			this.unsearchedBuildingIDs.clear();
			int clusterIndex = this.clustering.getClusterIndex(this.agentInfo.getID());
			Collection<StandardEntity> clusterEntities = this.clustering.getClusterEntities(clusterIndex);
			if (clusterEntities != null && clusterEntities.size() > 0) {
				for (StandardEntity entity : clusterEntities) {
					if (entity instanceof Building && entity.getStandardURN() != StandardEntityURN.REFUGE) {
						this.unsearchedBuildingIDs.add(entity.getID());
					}
				}
			} else {
				this.unsearchedBuildingIDs.addAll(this.worldInfo.getEntityIDsOfType(
						StandardEntityURN.BUILDING, 
						StandardEntityURN.GAS_STATION, 
						StandardEntityURN.AMBULANCE_CENTRE,
						StandardEntityURN.FIRE_STATION,
						StandardEntityURN.POLICE_OFFICE));
			}
		}
	}
	private boolean firstStage()
	{
		for(StandardEntity entity:this.clustering.getClusterEntities(this.SearchclusterIndex))
		{
			if(entity instanceof Building && ((Building) entity).isOnFire())
				return true;
		}
		return false;
	}
	private boolean secondStage()
	{
		for(EntityID entityid:this.worldInfo.getChanged().getChangedEntities())
		{
			if(entityid != null)
			{
				StandardEntity se = this.worldInfo.getEntity(entityid);
				if(se instanceof Building)
				{
					Building b = (Building) se;
					if(true && b.isFierynessDefined() && b.getFieryness() < 7)
						return false;
				}	
			}
		}
		return true;
	}
	@Override
	public EntityID getTarget() {
		return this.result;
	}
 	private void reflectMessage(MessageManager messageManager) {
	        Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
	        changedEntities.add(this.agentInfo.getID());
	        
	        for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
	            Class<? extends CommunicationMessage> messageClass = message.getClass();
	            if(messageClass == MessageBuilding.class) {
	                MessageBuilding mb = (MessageBuilding)message;
	                if(!changedEntities.contains(mb.getBuildingID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mb);
	                }
	            } else if(messageClass == MessageRoad.class) {
	                MessageRoad mr = (MessageRoad)message;
	                if(mr.isBlockadeDefined() && !changedEntities.contains(mr.getBlockadeID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mr);
	                }
	                
	            } else if(messageClass == MessageCivilian.class) {
	                MessageCivilian mc = (MessageCivilian) message;
	                if(!changedEntities.contains(mc.getAgentID())){
	                    MessageUtil.reflectMessage(this.worldInfo, mc);
	                }
	                
	            } else if(messageClass == MessageAmbulanceTeam.class) {
	                MessageAmbulanceTeam mat = (MessageAmbulanceTeam)message;
	                if(!changedEntities.contains(mat.getAgentID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mat);
	                }
	               
	            } else if(messageClass == MessageFireBrigade.class) {
	                MessageFireBrigade mfb = (MessageFireBrigade) message;
	                if(!changedEntities.contains(mfb.getAgentID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mfb);
	                }
	               
	            } else if(messageClass == MessagePoliceForce.class) {
	                MessagePoliceForce mpf = (MessagePoliceForce) message;
	                if(!changedEntities.contains(mpf.getAgentID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mpf);
	                }
	                
	            }
	        }
	    }

}