package MIC.module.complex;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.StandardMessagePriority;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageCivilian;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.complex.HumanDetector;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;



import static rescuecore2.standard.entities.StandardEntityURN.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import MIC.util.DistanceSorter;
import MIC.util.Agent.MIC_ATConfig;

public class MICHumanDetector extends HumanDetector {
    private Clustering clustering;
    protected PathPlanning pathPlanning;
    private EntityID result;

    private int sendTime;
    private int sendingAvoidTimeClearRequest;

    private Collection<EntityID> agentPositions;
    private Map<EntityID, Integer> sentTimeMap;
    private int sendingAvoidTimeReceived;
    private int sendingAvoidTimeSent;

    private int moveDistance;
    private EntityID lastPosition;
    private int positionCount;

    
    private ArrayList<StandardEntity> fire_bulding = new ArrayList<>();
    private ArrayList<StandardEntity> buldingCh = new ArrayList<>();
    
    private Civilian s ;

    private List<EntityID> visitedBuildingIDs;


    public MICHumanDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);

        this.result = null;
        this.sendTime = 0;
        
        s = null;
        
        this.sendingAvoidTimeClearRequest = developData.getInteger("MICHumanDetector.sendingAvoidTimeClearRequest", 5);

        this.agentPositions = new HashSet<>();
        this.sentTimeMap = new HashMap<>();
        visitedBuildingIDs = new ArrayList<EntityID>();
        this.sendingAvoidTimeReceived = developData.getInteger("MICHumanDetector.sendingAvoidTimeReceived", 3);
        this.sendingAvoidTimeSent = developData.getInteger("MICHumanDetector.sendingAvoidTimeSent", 5);

        this.moveDistance = developData.getInteger("MICHumanDetector.moveDistance", 40000);

        switch  (scenarioInfo.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.clustering = moduleManager.getModule("MICHumanDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                this.pathPlanning = moduleManager.getModule("MICHumanDetector.PathPlanning", "MIC.module.algorithm.MICPathPlanning");
                break;
            case PRECOMPUTED:
                this.clustering = moduleManager.getModule("MICHumanDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                this.pathPlanning = moduleManager.getModule("MICHumanDetector.PathPlanning", "MIC.module.algorithm.MICPathPlanning");
                break;
            case NON_PRECOMPUTE:
                this.clustering = moduleManager.getModule("MICHumanDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                this.pathPlanning = moduleManager.getModule("MICHumanDetector.PathPlanning", "MIC.module.algorithm.MICPathPlanning");
                break;
        }
    }

    @Override
    public HumanDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.clustering.updateInfo(messageManager);
        this.pathPlanning.updateInfo(messageManager);

        this.reflectMessage(messageManager);
        this.sendEntityInfo(messageManager);

        this.updateCivilianPosition();
        int currentTime = this.agentInfo.getTime();
        Human agent = (Human)this.agentInfo.me();
        int agentX = agent.getX();
        int agentY = agent.getY();
        StandardEntity positionEntity = this.worldInfo.getPosition(agent);
        if(positionEntity instanceof Road) {
            Road road = (Road)positionEntity;
            if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
                for(Blockade blockade : this.worldInfo.getBlockades(road)) {
                    if(blockade == null || !blockade.isApexesDefined()) {
                        continue;
                    }
                    if(MIC_ATConfig.isInsideApex(agentX, agentY, blockade.getApexes())) {
                            this.sendTime = currentTime;
                            messageManager.addMessage(
                                    new CommandPolice(
                                            true,
                                            StandardMessagePriority.HIGH,
                                            null,
                                            agent.getPosition(),
                                            CommandPolice.ACTION_CLEAR
                                    )
                            );
                            break;
                    }
                    /*if() {
                        this.sendTime = currentTime;
                        messageManager.addMessage(
                                new CommandPolice(
                                        true,
                                        StandardMessagePriority.HIGH,
                                        null,
                                        agent.getPosition(),
                                        CommandPolice.ACTION_CLEAR
                                )
                        );
                        break;
                }*/
                }
            }
            if(this.lastPosition != null && this.lastPosition.getValue() == road.getID().getValue()) {
                this.positionCount++;
                if(this.positionCount > this.getMaxTravelTime(road)) {
                    if ((this.sendTime + this.sendingAvoidTimeClearRequest) <= currentTime) {
                        this.sendTime = currentTime;
                    }
                }
            } else {
                this.lastPosition = road.getID();
                this.positionCount = 0;
            }
        }
        
        return this;
    }

    protected void updateCivilianPosition() {
        EntityID agentPositionID = agentInfo.getPosition();
        Collection<EntityID> humans = worldInfo.getEntityIDsOfType(StandardEntityURN.CIVILIAN);
        humans.addAll(worldInfo.getEntityIDsOfType(StandardEntityURN.AMBULANCE_TEAM));
        humans.addAll(worldInfo.getEntityIDsOfType(StandardEntityURN.POLICE_FORCE));
        humans.addAll(worldInfo.getEntityIDsOfType(StandardEntityURN.FIRE_BRIGADE));

        List<Human> humansToBeSaved = new ArrayList<Human>(humans.size());
        for (EntityID human : humans) {
            Human humanEntity = (Human)worldInfo.getEntity(human);
            if ((humanEntity.isBuriednessDefined() && humanEntity.getBuriedness() > 0) || (humanEntity.isDamageDefined() && humanEntity.getDamage() > 0)) {
                humansToBeSaved.add(humanEntity);
            }
        }
        List<EntityID> refugePath = pathPlanning.setFrom(agentInfo.getPosition()).setDestination(worldInfo.getEntityIDsOfType(StandardEntityURN.REFUGE)).calc().getResult();
        if (refugePath.size() > 0) {

            for (Human humanEntity : humansToBeSaved) {
                if (humanEntity.getPosition().equals(agentPositionID) && !worldInfo.getChanged().getChangedEntities().contains(humanEntity.getID())) {
                    humanEntity.setPosition(refugePath.get(refugePath.size() - 1));
                }
            }
        }
    }

    protected int stillWorkingCounter = 0;
    protected EntityID stillWorkingHuman = null;
    protected int stillWorkingThresh = 10;
    @Override
    
    public HumanDetector calc() 
    {
    	List<EntityID> changed = new ArrayList<EntityID>(this.worldInfo.getChanged().getChangedEntities());
    	for(EntityID c : changed)
    	{
    		if(this.worldInfo.getEntity(c) instanceof Building) buldingCh.add(this.worldInfo.getEntity(c));
    	}
    	for(StandardEntity m : buldingCh)
    	{
    		Building vc = (Building) m;
    		if(vc.isOnFire()) fire_bulding.add(vc);
    	}
    	
    	
        Human transportHuman =  this.agentInfo.someoneOnBoard();
        if(transportHuman != null) 
        {
            this.result = transportHuman.getID();
            return this;
        }

        boolean NRESULT = false;

        if(this.result != null && !(worldInfo.getPosition(result) instanceof AmbulanceTeam) && 
        		!(worldInfo.getPosition(result).getStandardURN() == REFUGE)) 
        {
            if (worldInfo.getPosition(result).getID().equals(agentInfo.getPosition())) 
            {
                if (!(agentInfo.getChanged().getChangedEntities().contains(result)))
                {
                	NRESULT = true;
                    this.result = null;
                } 
            }
            
            Human target = (Human) this.worldInfo.getEntity(this.result);
            if(target != null && target.getPosition().equals(agentInfo.getPosition()) && !NRESULT) 
            {
                if (!target.isHPDefined() || target.getHP() == 0) 
                {
                    this.result = null;
                } 
                else if (!target.isPositionDefined()) 
                {
                    this.result = null;
                } 
                else 
                {
                    StandardEntity position = this.worldInfo.getPosition(target);
                    if(position != null) 
                    {
                        StandardEntityURN positionURN = position.getStandardURN();
                        if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM) 
                        {
                            this.result = null;
                        }
                    }
                }
            }
        } 
        
        else if (agentInfo.getTime() > scenarioInfo.getKernelAgentsIgnoreuntil() && 
        		result != null && agentInfo.getPosition().equals(worldInfo.getPosition(result).getID()) && 
        		worldInfo.getPosition(agentInfo.getID()).getStandardURN() == REFUGE) 
        {
            this.result = null;

        }

        if(this.result == null) {
            if (clustering == null) {
                this.result = this.calcTargetInWorld();
                return this;
            }
            this.result = this.calcTargetInCluster(clustering);

            if (this.result == null) {
                this.result = this.calcTargetInWorld();
            }
        }
        if(fire_bulding.size() > 0 || fire_bulding != null || !(fire_bulding.isEmpty()))
        if(fire_bulding.contains(this.result))
        {
        	this.result = null;
        }
        return this;
    }

    private EntityID calcTargetInCluster(Clustering clustering) 
    {
        int clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
        Collection<StandardEntity> areas = clustering.getClusterEntities(clusterIndex);
        if(areas == null || areas.isEmpty()) 
        {
            return null;
        }

        List<Human> rescueTargets = new ArrayList<>();
        List<Human> loadTargets = new ArrayList<>();
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(AMBULANCE_TEAM, FIRE_BRIGADE, POLICE_FORCE)) 
        {
            Human h = (Human) next;
            StandardEntity positionEntity = this.worldInfo.getPosition(h);
            if(positionEntity != null && areas.contains(positionEntity)) 
            {
                if (h.isHPDefined() && h.isBuriednessDefined() && h.getHP() > 0 && h.getBuriedness() > 0) 
                {
                    boolean NotToAddTarget = false;
                    if (h.getPosition().equals(agentInfo.getPosition())) 
                    {
                        if (!agentInfo.getChanged().getChangedEntities().contains(h.getID())) 
                        {
                        	NotToAddTarget = true;
                        }
                    }
                    if (!NotToAddTarget) 
                    {
                        rescueTargets.add(h);
                    }
                }
            }
        }
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(CIVILIAN)) 
        {
            Human h = (Human) next;
            StandardEntity positionEntity = this.worldInfo.getPosition(h);
            if (positionEntity != null && areas.contains(positionEntity)) 
            {
                if (h.isHPDefined() && h.getHP() > 0) {
                    if (h.isBuriednessDefined() && h.getBuriedness() > 0) 
                    {
                        boolean doNotAddTarget = false;
                        if (h.getPosition().equals(agentInfo.getPosition())) 
                        {
                            if (!agentInfo.getChanged().getChangedEntities().contains(h.getID())) 
                            {
                                doNotAddTarget = true;
                            }
                        }
                        if (!doNotAddTarget) 
                        {
                            rescueTargets.add(h);
                        } 
                    } 
                    else 
                    {
                        if (h.isDamageDefined() && h.getDamage() > 0 && positionEntity.getStandardURN() != REFUGE) 
                        {
                            boolean doNotAddTarget = false;
                            if (h.getPosition().equals(agentInfo.getPosition())) 
                            {
                                if (!agentInfo.getChanged().getChangedEntities().contains(h.getID())) 
                                {
                                    doNotAddTarget = true;
                                }
                            }
                            if (!doNotAddTarget) 
                            {
                                loadTargets.add(h);
                            }
                        }
                        else 
                        {
                        	if (h.isDamageDefined() && h.isHPDefined() && h.getDamage() > 0 && positionEntity.getStandardURN() != REFUGE)
                        	{
                                boolean doNotAddTarget = false;
                        		if (h.isBuriednessDefined() && h.getBuriedness() >= 0)
                        		{
                        			if (!agentInfo.getChanged().getChangedEntities().contains(h.getID())) 
                                    {
                                        doNotAddTarget = true;
                                    }
                        			if (!doNotAddTarget) 
                                    {
                        				if(h instanceof Civilian)
                        					loadTargets.add(h);
                                    }
                        		}
                        	}
                        }
                    }
                }
            }
        }
        if(rescueTargets.size() > 0) {
            rescueTargets.sort(new DistanceSorter(this.worldInfo, this.agentInfo.me()));
            return rescueTargets.get(0).getID();
        }
        if(loadTargets.size() > 0) {
            loadTargets.sort(new DistanceSorter(this.worldInfo, this.agentInfo.me()));
            return loadTargets.get(0).getID();
        }
        return null;
    }

    private EntityID calcTargetInWorld() 
    {
    	
        List<Human> rescueTargets = new ArrayList<>();
        List<Human> loadTargets = new ArrayList<>();
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(AMBULANCE_TEAM, FIRE_BRIGADE, POLICE_FORCE)) 
        {
            Human h = (Human) next;
            if (this.agentInfo.getID().getValue() != h.getID().getValue()) 
            {
                StandardEntity positionEntity = this.worldInfo.getPosition(h);
                if (positionEntity != null && h.isHPDefined() && h.isBuriednessDefined()) 
                {
                    if (h.isHPDefined() && h.getHP() > 0 && h.getBuriedness() > 0) 
                    {
                        rescueTargets.add(h);
                    }
                }
            }
        }
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(CIVILIAN)) 
        {
            Human h = (Human) next;
            StandardEntity positionEntity = this.worldInfo.getPosition(h);
            if(positionEntity != null && positionEntity instanceof Area) 
            {
                if (h.isHPDefined() && h.getHP() > 0) 
                {
                    if (h.isBuriednessDefined() && h.getHP() < 10000 &&h.isDamageDefined() &&(h.getBuriedness() >= 0 || h.getDamage() > 0)&& positionEntity.getStandardURN() != REFUGE) 
                    {
                        rescueTargets.add(h);
                    } 
                    else 
                    {
                        if (h.isDamageDefined() && h.getDamage() > 0  && positionEntity.getStandardURN() != REFUGE) 
                        {
                            loadTargets.add(h);
                        }
                    }
                }
            }
        }

        Collection<StandardEntity> agents = worldInfo.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM);
        List<AmbulanceTeam> ambulanceAgents = new ArrayList<>();
        for (StandardEntity se : agents) 
        {
            AmbulanceTeam ambulance = (AmbulanceTeam) se;
            if (ambulance.isHPDefined() && ambulance.getHP() > 0 && ambulance.getBuriedness() == 0 && ambulance.isPositionDefined() && ambulance.getPosition().equals(agentInfo.getPosition())) 
            {
                ambulanceAgents.add(ambulance);
            }
        }

        while(rescueTargets.size() > 0) 
        {
            for (int i = 0; i < ambulanceAgents.size(); i++) 
            {
                rescueTargets.sort(new DistanceSorter(this.worldInfo, ambulanceAgents.get(i)));
                if (ambulanceAgents.get(i).getID().equals(agentInfo.getID())) 
                {
                    return rescueTargets.get(0).getID();
                }
                rescueTargets.remove(0);
                if (rescueTargets.size() == 0)
                    break;
            }
        }
        while(loadTargets.size() > 0) 
        {
            for (int i = 0; i < ambulanceAgents.size(); i++) 
            {
                loadTargets.sort(new DistanceSorter(this.worldInfo,ambulanceAgents.get(i)));
                if (ambulanceAgents.get(i).getID().equals(agentInfo.getID())) 
                {
                    return loadTargets.get(0).getID();
                }
                loadTargets.remove(0);
                if (loadTargets.size() == 0)
                    break;
            }
        }
        return null;
    }

    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public HumanDetector precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.clustering.precompute(precomputeData);
        return this;
    }

    @Override
    public HumanDetector resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.clustering.resume(precomputeData);
        preparate();
        return this;
    }

    @Override
    public HumanDetector preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.clustering.preparate();
        this.pathPlanning.preparate();
        return this;
    }

    private void reflectMessage(MessageManager messageManager) 
    {
        
        List<EntityID> changedEntities = new ArrayList<EntityID>(this.worldInfo.getChanged().getChangedEntities());
        Human agent = (Human) this.agentInfo.me();
        changedEntities.add(this.agentInfo.getID());
        int time = this.agentInfo.getTime();
        for(CommunicationMessage message : messageManager.getReceivedMessageList()) 
        {
            Class<? extends CommunicationMessage> messageClass = message.getClass();
        	if(!(agent instanceof AmbulanceTeam)) 
            if(messageClass == MessageCivilian.class) 
            {
            	
                MessageCivilian mc = (MessageCivilian) message;
                if(!changedEntities.contains(mc.getAgentID()))
                {
                    MessageUtil.reflectMessage(this.worldInfo, mc);
                }
                this.sentTimeMap.put(mc.getAgentID(), time + this.sendingAvoidTimeReceived);
            } 
            else if(messageClass == MessageAmbulanceTeam.class) 
            {
                MessageAmbulanceTeam mat = (MessageAmbulanceTeam)message;
                if(!changedEntities.contains(mat.getAgentID())) 
                {
                    MessageUtil.reflectMessage(this.worldInfo, mat);
                }
                this.sentTimeMap.put(mat.getAgentID(), time + this.sendingAvoidTimeReceived);
            } 
            else if(messageClass == MessageFireBrigade.class) 
            {
                MessageFireBrigade mfb = (MessageFireBrigade) message;
                if(!changedEntities.contains(mfb.getAgentID())) 
                {
                    MessageUtil.reflectMessage(this.worldInfo, mfb);
                }
                this.sentTimeMap.put(mfb.getAgentID(), time + this.sendingAvoidTimeReceived);
            } 
            else if(messageClass == MessagePoliceForce.class) 
            {
                MessagePoliceForce mpf = (MessagePoliceForce) message;
                if(!changedEntities.contains(mpf.getAgentID())) 
                {
                    MessageUtil.reflectMessage(this.worldInfo, mpf);
                }
                this.sentTimeMap.put(mpf.getAgentID(), time + this.sendingAvoidTimeReceived);
            }
        }
    }

    private boolean checkSendFlags(){
        boolean isSendCivilianMessage = true;

        StandardEntity me = this.agentInfo.me();
        if(!(me instanceof Human)){
            return false;
        }
        Human agent = (Human)me;
        EntityID agentID = agent.getID();
        EntityID position = agent.getPosition();
        StandardEntityURN agentURN = agent.getStandardURN();
        EnumSet<StandardEntityURN> agentTypes = EnumSet.of(AMBULANCE_TEAM, FIRE_BRIGADE, POLICE_FORCE);
        agentTypes.remove(agentURN);

        this.agentPositions.clear();
        for(StandardEntity entity : this.worldInfo.getEntitiesOfType(agentURN)) {
            Human other = (Human)entity;
            if(isSendCivilianMessage) {
                if (other.getPosition().getValue() == position.getValue()) {
                    if (other.getID().getValue() > agentID.getValue()) {
                        isSendCivilianMessage = false;
                    }
                }
            }
            this.agentPositions.add(other.getPosition());
        }

        for(StandardEntityURN urn : agentTypes) {
            for(StandardEntity entity : this.worldInfo.getEntitiesOfType(urn)) {
                Human other = (Human)entity;
                if(isSendCivilianMessage) {
                    if (other.getPosition().getValue() == position.getValue()) {
                        if (urn == AMBULANCE_TEAM) {
                            isSendCivilianMessage = false;
                        } else if(agentURN != AMBULANCE_TEAM && other.getID().getValue() > agentID.getValue()) {
                            isSendCivilianMessage = false;
                        }
                    }
                }
                this.agentPositions.add(other.getPosition());
            }
        }
        return isSendCivilianMessage;
    }

    private void sendEntityInfo(MessageManager messageManager) {
        if(this.checkSendFlags()) {
            Civilian civilian = null;
            int currentTime = this.agentInfo.getTime();
            Human agent = (Human) this.agentInfo.me();
            for (EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
                StandardEntity entity = Objects.requireNonNull(this.worldInfo.getEntity(id));
                if (entity.getStandardURN() == CIVILIAN) {
                    Integer time = this.sentTimeMap.get(id);
                    if (time != null && time > currentTime) {
                        continue;
                    }
                    Civilian target = (Civilian) entity;
                    if (!this.agentPositions.contains(target.getPosition())) {
                        civilian = this.selectCivilian(civilian, target);
                    } else if (target.getPosition().getValue() == agent.getPosition().getValue()) {
                        civilian = this.selectCivilian(civilian, target);
                    }
                }
            }
            if (civilian != null) {
                messageManager.addMessage(new MessageCivilian(true, civilian));
                this.sentTimeMap.put(civilian.getID(), currentTime + this.sendingAvoidTimeSent);
            }
        }
    }

    private Civilian selectCivilian(Civilian civilian1, Civilian civilian2) {
        if(this.checkCivilian(civilian1, true)) {
            if(this.checkCivilian(civilian2, true)) {
                if(civilian1.getHP() > civilian2.getHP()) {
                    return civilian1;
                } else if(civilian1.getHP() < civilian2.getHP()) {
                    return civilian2;
                }else {
                    if(civilian1.isBuriednessDefined() && civilian2.isBuriednessDefined()) {
                        if(civilian1.getBuriedness() > 0 && civilian2.getBuriedness() == 0) {
                            return civilian1;
                        }else if(civilian1.getBuriedness() == 0 && civilian2.getBuriedness() > 0) {
                            return civilian2;
                        } else {
                            if (civilian1.getBuriedness() < civilian2.getBuriedness()) {
                                return civilian1;
                            } else if (civilian1.getBuriedness() > civilian2.getBuriedness()) {
                                return civilian2;
                            }
                        }
                    }
                    if(civilian1.isDamageDefined() && civilian2.isDamageDefined()) {
                        if(civilian1.getDamage() < civilian2.getDamage()) {
                            return civilian1;
                        } else if(civilian1.getDamage() > civilian2.getDamage()) {
                            return civilian2;
                        }
                    }
                }
            }
            return civilian1;
        }
        if(this.checkCivilian(civilian2, true)) {
            return civilian2;
        } else if(this.checkCivilian(civilian1, false)) {
            return civilian1;
        } else if(this.checkCivilian(civilian2, false)) {
            return civilian2;
        }
        return null;
    }

    private boolean checkCivilian(Civilian c, boolean checkOtherValues) {
        if(c != null && c.isHPDefined() && c.isPositionDefined()) {
            if(checkOtherValues && (!c.isDamageDefined() && !c.isBuriednessDefined())) {
                return false;
            }
            return true;
        }
        return false;
    }
    private int getMaxTravelTime(Area area) {
        int distance = 0;
        List<Edge> edges = new ArrayList<>();
        for(Edge edge : area.getEdges()) {
            if(edge.isPassable()) {
                edges.add(edge);
            }
        }
        if(edges.size() <= 1) {
            return Integer.MAX_VALUE;
        }
        for(int i = 0; i < edges.size(); i++) {
            for(int j = 0; j < edges.size(); j++) {
                if(i != j) {
                    Edge edge1 = edges.get(i);
                    double midX1 = (edge1.getStartX() + edge1.getEndX()) / 2;
                    double midY1 = (edge1.getStartY() + edge1.getEndY()) / 2;
                    Edge edge2 = edges.get(j);
                    double midX2 = (edge2.getStartX() + edge2.getEndX()) / 2;
                    double midY2 = (edge2.getStartY() + edge2.getEndY()) / 2;
                    int d = this.getDistance(midX1, midY1, midX2, midY2);
                    if(distance < d) {
                        distance = d;
                    }
                }
            }
        }
        if(distance > 0) {
            return (distance / this.moveDistance) + ((distance % this.moveDistance) > 0 ? 1 : 0) + 1;
        }
        return Integer.MAX_VALUE;
    }
    
    private int getDistance(double fromX, double fromY, double toX, double toY) {
        double dx = toX - fromX;
        double dy = toY - fromY;
        return (int)Math.hypot(dx, dy);
    }
}