package MIC.module.complex;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.communication.standard.bundle.information.MessageRoad;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.algorithm.StaticClustering;
import adf.component.module.complex.RoadDetector;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import MIC.util.CollectionUtil;
import MIC.util.Agent.MIC_PFConfig;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class MICRoadDetector extends RoadDetector {
	
	private MIC_PFConfig policeConfig;
	
    private Set<EntityID> targetAreas;
    private Set<EntityID> priorityRoads;
    private PathPlanning pathPlanning;
    private StaticClustering clustering;
    private EntityID result;
    
    
	private Boolean refugeMode = true;
	private List<StandardEntity> policeAgents;
	private List<StandardEntity> refuges;
	private List<StandardEntity> chosenResults;
	private List<StandardEntity> visited;
	private int refugeModeCounter = 0;
	private StandardEntity refugeRoad = null;

	private CollectionUtil collectionUtil;

	private List<EntityID> visitedRoads = new ArrayList<EntityID>();;
	
    private List<CommandPolice> receivedMessages = new ArrayList<CommandPolice>();

	
    private EntityID toughTarget = null;
    private Set<EntityID> clusterEntitiesSet;
    private EntityID visitedPosition;
        

    public MICRoadDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        switch (scenarioInfo.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("MICRoadDetector.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.clustering = moduleManager.getModule("MICRoadDetector.Clustering", "MIC.module.algorithm.MICKMeans");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("MICRoadDetector.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.clustering = moduleManager.getModule("MICRoadDetector.Clustering", "MIC.module.algorithm.MICKMeans");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("MICRoadDetector.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.clustering = moduleManager.getModule("MICRoadDetector.Clustering", "MIC.module.algorithm.MICKMeans");
                break;
        }
        this.result = null;
        visitedPosition = null;
        refuges = new ArrayList<StandardEntity>(this.worldInfo.getEntitiesOfType(REFUGE));
        policeAgents = new ArrayList<StandardEntity>(this.worldInfo.getEntitiesOfType(POLICE_FORCE));
        priorityRoads = new HashSet<>();
        chosenResults = new ArrayList<>();
        visited = new ArrayList<>();
        collectionUtil = new CollectionUtil();
        policeConfig = new MIC_PFConfig(this.worldInfo, this.agentInfo, this.pathPlanning, this.scenarioInfo);
    }

	
    @Override
    public RoadDetector calc() {
    	
    	updateVisited(visitedPosition, agentInfo.getPosition());

    	if(this.refugeMode)
		{
			if(policeAgents.size() >= 5)
			{
				if(agentInfo.getID().getValue() == policeAgents.get(4).getID().getValue())
				{
					if(agentInfo.getID().getValue() == policeAgents.get(4).getID().getValue())
					{
						StandardEntity myResult = null;
				    	if(chosenResults.isEmpty() || chosenResults == null || chosenResults.size() == 0)
						{
				    		this.refugeModeCounter += 1;
				    		chosenResults.addAll(policeConfig.detectRefugeEntrance( refuges.get(0), this.worldInfo, pathPlanning ));		
				    		visited.addAll(chosenResults);
							StandardEntity repeatEntity = refuges.get(0);
							collectionUtil.removeItemFromList(refuges, refuges.get(0));
							
							refuges.add(repeatEntity);
							myResult = chosenResults.get(0);
							this.refugeRoad = myResult;
							collectionUtil.removeItemFromList(chosenResults, chosenResults.get(0));
							
							if(this.refugeModeCounter > refuges.size())
							{
								this.refugeMode = false;
							}
						}
				    	else
				    	{
				    		Road rr = (Road) refugeRoad;
				    		
				    		if(rr != null && rr.isBlockadesDefined())
				    		{
				    			if(rr.getBlockades().size() == 0)
				    			{
									myResult = chosenResults.get(0);
									collectionUtil.removeItemFromList(chosenResults, chosenResults.get(0));
				    			}
				    		}
				    		else if(!rr.isBlockadesDefined() && rr != null)
							{
								myResult = this.refugeRoad;
							}
				    	}

				    	if(myResult != null)
				    	{
				    		this.result = myResult.getID();
				    		this.pathPlanning.setFrom(this.agentInfo.getPosition());
				    		this.pathPlanning.setDestination(this.result);
				    		return this;
				    	}
				    	else
				    	{
				    		this.result = this.refugeRoad.getID();
				    		this.pathPlanning.setFrom(this.agentInfo.getPosition());
				    		this.pathPlanning.setDestination(this.result);
				    		return this;
				    	}
					}
				}
			}
		}
    	this.result = policeConfig.calcCommunication(receivedMessages);
    	if(this.result == null)
    		this.result = policeConfig.calcStuckHuman();
		if (this.result == null)
			this.result = policeConfig.calcCommon(this.priorityRoads, this.targetAreas);
        if (toughTarget != null)
        {
            this.result = toughTarget;
            visitedPosition = agentInfo.getPosition();
            return this;
        }
        /*
        EntityID agentPosition = agentInfo.getPosition();
        pathPlanning.setFrom(agentPosition);
        pathPlanning.setDestination(clusterEntitiesSet);
        List<EntityID> path = pathPlanning.calc().getResult();
        if (path != null && path.size() > 0)
        {
            this.result = path.get(path.size()-1);
            toughTarget = result;
            visitedPosition = agentInfo.getPosition();
            return this;
        }
        visitedPosition = agentInfo.getPosition();*/
        return this;
    }

    private void updateVisited(EntityID start, EntityID target) {
        if (target == null || start == null)
            return;

        visitedRoads.add(start);
        
        List<EntityID> path = pathPlanning.getResult(start,target);
        if (path != null) {
            for (EntityID entity : path) {
                visitedRoads.add(entity);
            }
        }
    }

    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public RoadDetector precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if (this.getCountPrecompute() >= 2) {
            return this;
        }
        return this;
    }

    @Override
    public RoadDetector resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if (this.getCountResume() >= 2) {
            return this;
        }
        preparate();
        return this;
    }

    @Override
    public RoadDetector preparate() {
        super.preparate();
        if (this.getCountPreparate() >= 2) {
            return this;
        }
        
        this.pathPlanning.preparate();
        this.targetAreas = new HashSet<>();
        this.targetAreas.addAll(policeConfig.calcPreparate());
        return this;
    }

    @Override
    public RoadDetector updateInfo(MessageManager messageManager) {
    	
    	
        super.updateInfo(messageManager);
        if (this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.pathPlanning.updateInfo(messageManager);
        if (this.result != null) {
            if (this.agentInfo.getPosition().equals(this.result)) 
            {
                StandardEntity entity = this.worldInfo.getEntity(this.result);
                if (entity instanceof Building)
                {
                    this.result = null;
                } 
                else if (entity instanceof Road)
                {
                    Road road = (Road) entity;
                    if (!road.isBlockadesDefined() || road.getBlockades().isEmpty())
                    {
                        this.targetAreas.remove(this.result);
                        this.result = null;
                    }
                }
            }
        }
        
        Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
        
        for (CommunicationMessage message : messageManager.getReceivedMessageList()) {
            Class<? extends CommunicationMessage> messageClass = message.getClass();
            if (messageClass == MessageAmbulanceTeam.class)
            {
                this.reflectMessage((MessageAmbulanceTeam) message);
            }
            
            else if (messageClass == MessageFireBrigade.class) 
            {
                this.reflectMessage((MessageFireBrigade) message);
            }
            
            else if (messageClass == MessageRoad.class)
            {
                this.reflectMessage((MessageRoad) message, changedEntities);
            }

            else if (messageClass == MessagePoliceForce.class) 
            {
                this.reflectMessage((MessagePoliceForce) message);
            }

        }
        
        
        for (EntityID id : this.worldInfo.getChanged().getChangedEntities())
        {
            StandardEntity entity = this.worldInfo.getEntity(id);
            if (entity instanceof Road)
            {
                Road road = (Road) entity;
                if (!road.isBlockadesDefined() || road.getBlockades().isEmpty())
                {
                    this.targetAreas.remove(id);
                }
            }
        }

        this.receivedMessages = policeConfig.gatherMessages(messageManager);        
        return this;
    }

    private void reflectMessage(MessageRoad messageRoad, Collection<EntityID> changedEntities) {
        if (messageRoad.isBlockadeDefined() && !changedEntities.contains(messageRoad.getBlockadeID())) {
            MessageUtil.reflectMessage(this.worldInfo, messageRoad);
        }
        if (messageRoad.isPassable()) {
            this.targetAreas.remove(messageRoad.getRoadID());
        }
    }

    private void reflectMessage(MessageAmbulanceTeam messageAmbulanceTeam) {
        if (messageAmbulanceTeam.getPosition() == null) {
            return;
        }
        if (messageAmbulanceTeam.getAction() == MessageAmbulanceTeam.ACTION_RESCUE) 
        {
            StandardEntity position = this.worldInfo.getEntity(messageAmbulanceTeam.getPosition());
            if (position != null && position instanceof Building)
            {
                this.targetAreas.removeAll(((Building) position).getNeighbours());
            }
        }
        else if (messageAmbulanceTeam.getAction() == MessageAmbulanceTeam.ACTION_LOAD) 
        {
            StandardEntity position = this.worldInfo.getEntity(messageAmbulanceTeam.getPosition());
            if (position != null && position instanceof Building)
            {
                this.targetAreas.removeAll(((Building) position).getNeighbours());
            }
        } 
        else if (messageAmbulanceTeam.getAction() == MessageAmbulanceTeam.ACTION_MOVE)
        {
            if (messageAmbulanceTeam.getTargetID() == null)
            {
                return;
            }
            StandardEntity target = this.worldInfo.getEntity(messageAmbulanceTeam.getTargetID());
            if (target instanceof Building)
            {
                for (EntityID id : ((Building) target).getNeighbours())
                {
                    StandardEntity neighbour = this.worldInfo.getEntity(id);
                    if (neighbour instanceof Road)
                    {
                        this.priorityRoads.add(id);
                    }
                }
            }
            else if (target instanceof Human)
            {
                Human human = (Human) target;
                if (human.isPositionDefined())
                {
                    StandardEntity position = this.worldInfo.getPosition(human);
                    if (position instanceof Building)
                    {
                        for (EntityID id : ((Building) position).getNeighbours())
                        {
                            StandardEntity neighbour = this.worldInfo.getEntity(id);
                            if (neighbour instanceof Road)
                            {
                                this.priorityRoads.add(id);
                            }
                        }
                    }
                }
            }
        }
    }

    private void reflectMessage(MessageFireBrigade messageFireBrigade) {
    	if (messageFireBrigade.getTargetID() == null)
        {
            return;
        }
        if (messageFireBrigade.getAction() == MessageFireBrigade.ACTION_REFILL)
        {
            StandardEntity target = this.worldInfo.getEntity(messageFireBrigade.getTargetID());
            if (target instanceof Building)
            {
                for (EntityID id : ((Building) target).getNeighbours())
                {
                    StandardEntity neighbour = this.worldInfo.getEntity(id);
                    if (neighbour instanceof Road)
                    {
                        this.priorityRoads.add(id);
                    }
                }
            }
            else if (target.getStandardURN() == HYDRANT)
            {
                this.priorityRoads.add(target.getID());
                this.targetAreas.add(target.getID());
            }
        }
    }

    private void reflectMessage(MessagePoliceForce messagePoliceForce) {
    	if (messagePoliceForce.getAction() == MessagePoliceForce.ACTION_CLEAR)
        {
            if (messagePoliceForce.getAgentID().getValue() != this.agentInfo.getID().getValue())
            {
                if (messagePoliceForce.isTargetDefined())
                {
                    EntityID targetID = messagePoliceForce.getTargetID();
                    if (targetID == null)
                    {
                        return;
                    }
                    StandardEntity entity = this.worldInfo.getEntity(targetID);
                    if (entity == null)
                    {
                        return;
                    }

                    if (entity instanceof Area)
                    {
                        this.targetAreas.remove(targetID);
                        if (this.result != null && this.result.getValue() == targetID.getValue())
                        {
                            if (this.agentInfo.getID().getValue() < messagePoliceForce.getAgentID().getValue())
                            {
                                this.result = null;
                            }
                        }
                    }
                    else if (entity.getStandardURN() == BLOCKADE)
                    {
                        EntityID position = ((Blockade) entity).getPosition();
                        this.targetAreas.remove(position);
                        if (this.result != null && this.result.getValue() == position.getValue())
                        {
                            if (this.agentInfo.getID().getValue() < messagePoliceForce.getAgentID().getValue())
                            {
                                this.result = null;
                            }
                        }
                    }

                }
            }
        }
    }
}