package MIC.util.Agent;
import static rescuecore2.standard.entities.StandardEntityURN.BUILDING;
import static rescuecore2.standard.entities.StandardEntityURN.GAS_STATION;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import MIC.util.Convertors;
import MIC.util.DistanceSorter;
import MIC.util.CollectionUtil;
import MIC.util.RandomGenerator;
import adf.agent.info.AgentInfo;
import adf.agent.info.WorldInfo;
import adf.component.module.algorithm.Clustering;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;


public class MIC_FBConfig {

	private Convertors convertor;
	private CollectionUtil collectionUtil;
	private RandomGenerator MICRandom;
	private WorldInfo worldInfo;
	public MIC_FBConfig(WorldInfo wi)
	{
		convertor = new Convertors(wi);
		collectionUtil = new CollectionUtil();
		MICRandom = new RandomGenerator();
		this.worldInfo = wi;
	}
	
	public EntityID firstStage(WorldInfo wi, List<StandardEntity> fireAgents, AgentInfo agentInfo)
	{
		EntityID result = null;
		
		int range = 3;
		
		List<StandardEntity> changedEntities = convertor.getStandardEntityList( new ArrayList<EntityID>(wi.getChanged().getChangedEntities()));
		List<StandardEntity> remove = new ArrayList<>();
		
		
		for(StandardEntity entity : changedEntities)
		{
			if(!( entity instanceof Building ))
				remove.add(entity);
			else
			{
				Building building = (Building) entity;
				if( !(building.isOnFire() && building.isFierynessDefined()) )
					remove.add(entity);
			}
		}
		
		changedEntities.removeAll(remove);
		
		List<List<StandardEntity>> priorityLevels = new ArrayList<List<StandardEntity>>();
		
		for(int level = 0 ; level < range ; level++)
		{
			int fieryness = level + 1;
			List<StandardEntity> currentLevel = new ArrayList<StandardEntity>();
			for(StandardEntity entity : changedEntities)
			{
				Building building = (Building) entity;
				if(building.getFieryness() == fieryness && currentLevel.contains(entity) == false)
					currentLevel.add(entity);
			}
			priorityLevels.add(level, currentLevel);
		}
		
		if(priorityLevels.size() == 0) return null;
		
		for(int level = 0 ; level < range ; level++)
		{
			List<StandardEntity> current = priorityLevels.get(level);
			if(current.size() > 0)
			{
				current.sort(new DistanceSorter(wi, agentInfo.me()));
				if(current.size() == 1)
					result = current.get(0).getID();
				else
					result = current.get(this.MICRandom.GenerateRandomNumber(0, 1)).getID();
				
			}
			else
				continue;
			
		}
		return result;
	}
	
	 public EntityID secondStage(WorldInfo wi, AgentInfo agentInfo) {
	        Collection<StandardEntity> entities = wi.getEntitiesOfType(
	                BUILDING,
	                GAS_STATION
	        );
	        StandardEntity me = agentInfo.me();
	        
	        Set<StandardEntity> fireBuildings = new HashSet<>();
	        for (StandardEntity entity : entities) {
	            if (((Building)entity).isOnFire()) {
	                fireBuildings.add(entity);
	            }
	        }

	        List<StandardEntity> agents = new ArrayList<>(wi.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE));

	        
	        for(StandardEntity entity : fireBuildings) {
	            if(agents.isEmpty()) {
	                break;
	            } else if(agents.size() == 1) {
	                if(agents.get(0).getID().getValue() == me.getID().getValue()) {
	                    return entity.getID();
	                }
	                break;
	            }
	            agents.sort(new DistanceSorter(wi, entity));
	            StandardEntity a0 = agents.get(0);
	            StandardEntity a1 = agents.get(1);

	            if(me.getID().getValue() == a0.getID().getValue() || me.getID().getValue() == a1.getID().getValue()) {
	                return entity.getID();
	            } else {
	                agents.remove(a0);
	                agents.remove(a1);
	            }
	        }
	        return null;
	    }

    public List<StandardEntity> detectRefugeBuildings(WorldInfo wi, StandardEntity refuge)
    {
    	
    	List<StandardEntity> FPBuildings = new ArrayList<StandardEntity>();
    	List<EntityID> buildings = selectNearbyBuildings (wi, refuge.getID());
		Map<StandardEntity, Integer> BuildingDistances = new HashMap<StandardEntity, Integer>();
			
		for(EntityID building : buildings)
		{
			
			StandardEntity standardBuilding = wi.getEntity(building);
    		
			if(building != null)
			{   
				int distance = wi.getDistance(building, refuge.getID());
				BuildingDistances.put(standardBuilding, distance);
    		}
		}
		
		Map<StandardEntity, Integer> sorted = collectionUtil.MapSorter(BuildingDistances, true);
		List<StandardEntity> finalList = new ArrayList<StandardEntity>(sorted.keySet());
		
		for(Iterator<StandardEntity> iterator = finalList.iterator(); iterator.hasNext();)
		{        		
			StandardEntity object = iterator.next();
			if(object != finalList.get(0)
					&& object != finalList.get(1)
					&& object != finalList.get(2)
					&& object != finalList.get(3))
				iterator.remove();
		}
		
		FPBuildings.addAll(finalList);		
		return FPBuildings;
    }
    
    public Boolean needBreak(WorldInfo wi, List<EntityID> neighbours)
    {
    	int count = 0;
    	for(EntityID se : neighbours)
    	{
    		StandardEntity e = wi.getEntity(se);
    		if(e instanceof Building)
    			count += 1;
    	}
    	if(count == 0) return true;
    	else return false;
    }
    
   
    
    List<EntityID> repeatedNeighbours = new ArrayList<EntityID>();
    
    public List<StandardEntity> detectRefugeNeighbour(WorldInfo wi, StandardEntity refuge)
    { 
    	
    	List<StandardEntity> result = new ArrayList<StandardEntity>();
    	Building b = (Building) refuge;
    	List<EntityID> neighbours = b.getNeighbours();
    	for(EntityID eid : neighbours)
    	{
    		if(wi.getEntity(eid) instanceof Building)
    		{
    			ArrayList<EntityID> ce = (ArrayList<EntityID>) (((Building) wi.getEntity(eid)).getNeighbours());
    			repeatedNeighbours = collectionUtil.removeIDListFromList(ce, repeatedNeighbours);
    			if(needBreak(wi, ce))
    			{
    				result.add(wi.getEntity(eid));
    				continue;
    			}
    			repeatedNeighbours.add(eid);
    			result.addAll(this.detectRefugeNeighbour(wi, wi.getEntity(eid) ));
    			result.add(wi.getEntity(eid));
    			
    		}
    		else
    			continue;
    	}
    	
    	return result;
    	
    } 

    public boolean isInsideApex(double pX, double pY, int[] apex) {
        Point2D p = new Point2D(pX, pY);
        Vector2D v1 = (new Point2D(apex[apex.length - 2], apex[apex.length - 1])).minus(p);
        Vector2D v2 = (new Point2D(apex[0], apex[1])).minus(p);
        double theta = getAngle(v1, v2);

        for(int i = 0; i < apex.length - 2; i += 2) {
            v1 = (new Point2D(apex[i], apex[i + 1])).minus(p);
            v2 = (new Point2D(apex[i + 2], apex[i + 3])).minus(p);
            theta += getAngle(v1, v2);
        }
        return Math.round(Math.abs((theta / 2) / Math.PI)) >= 1;
    }

    private static double getAngle(Vector2D v1, Vector2D v2) {
        double flag = (v1.getX() * v2.getY()) - (v1.getY() * v2.getX());
        double angle = Math.acos(((v1.getX() * v2.getX()) + (v1.getY() * v2.getY())) / (v1.getLength() * v2.getLength()));
        if(flag > 0) {
            return angle;
        }
        if(flag < 0) {
            return -1 * angle;
        }
        return 0.0D;
    }

	public List<EntityID> selectNearbyBuildings(WorldInfo wi, EntityID Refuge)
	{
		List<EntityID> nearby = new ArrayList<EntityID>();
		List<EntityID> nearbyBuildings = new ArrayList<EntityID>();
		
		int range = 200;
	
		while(nearbyBuildings.size() < 20)
		{
			
			nearby = new ArrayList<EntityID>(wi.getObjectIDsInRange(Refuge, range));
		
			for(EntityID ei : nearby)
			{
				StandardEntity se = wi.getEntity(ei);
				if ( nearbyBuildings.contains(ei) )
					continue;
				
				List<StandardEntity> buildings = new ArrayList<StandardEntity>(wi.getEntitiesOfType(BUILDING));
				if(buildings.contains( se ))
				{
					nearbyBuildings.add( se.getID() );
				}
			}
			if(nearbyBuildings.size() < 20)
				range += 100;	
		}
		
		return nearbyBuildings;
	}
	
	
	 public Collection<EntityID> removeWorthlessTargets(Collection<EntityID> unsearched, AgentInfo agentInfo)
	   {
		   unsearched = collectionUtil.removeItemFromList(new ArrayList<EntityID>(unsearched), agentInfo.getPosition());
		   for (EntityID id : agentInfo.getChanged().getChangedEntities())
		   {
	           StandardEntity entity = this.worldInfo.getEntity(id);
	           if (entity instanceof Building)
	           {
	        	   Building b = (Building) entity;
	        	   
	        	   if (b.isFierynessDefined() && b.getFieryness() > 4)
	        	   {
	        		   unsearched = collectionUtil.removeItemFromList(new ArrayList<EntityID>(unsearched), id);   
	        	   }
	           }
	       }
		   
		   List<EntityID> othersPosotion = new ArrayList<EntityID>();
		   othersPosotion.addAll(this.getAgentURNPositions(StandardEntityURN.FIRE_BRIGADE));
		   unsearched = collectionUtil.removeIDListFromList(new ArrayList<EntityID>(unsearched), othersPosotion);
		   
		   return unsearched;
	   }
	   
	   public List<EntityID> getAgentURNPositions(StandardEntityURN urn)
	   {
		   List<EntityID> positions = new ArrayList<EntityID>();
		   List<StandardEntity> agents = new ArrayList<StandardEntity>(this.worldInfo.getEntitiesOfType(urn));
		   
		   if(urn == StandardEntityURN.FIRE_BRIGADE)
		   {
			   for(StandardEntity entity : agents)
			   {
				   if (entity instanceof AmbulanceTeam) {
		               AmbulanceTeam currentAgent = (AmbulanceTeam) entity;
		               if ( currentAgent.isPositionDefined() && currentAgent.getPosition() != null )
		               {
		                   if ( positions.contains(currentAgent.getPosition()) == false )
		                	   positions.add(currentAgent.getPosition());
		               }
		           }
			   }
		   }
		   return positions;
	   }
	   
	   public Collection<EntityID> customSearchReset(Collection<EntityID> unsearched, Clustering clustering, int index)
	   {
		   unsearched.clear();
		   Collection<StandardEntity> clusterEntities = new ArrayList<>();
		   index -=1 ;
		   do
		   {
			   index += 1;
			   clusterEntities = clustering.getClusterEntities(index);
		   }
		   while( clusterEntities.isEmpty() && index < clustering.getClusterNumber() );
	       
		   if (clusterEntities != null && clusterEntities.size() > 0)
	       {
	           for (StandardEntity entity : clusterEntities) {
	               if (entity instanceof Building)
	               {
	            	   Building b = (Building) entity;
	            	   if(b.isOnFire())
					   {
	            		   unsearched.add(entity.getID());
					   }
	               }
	           }
	       }
	       return unsearched;
	   }
}