package MIC.util.Agent;
import static rescuecore2.standard.entities.StandardEntityURN.BLOCKADE;
import static rescuecore2.standard.entities.StandardEntityURN.BUILDING;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.GAS_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import static rescuecore2.standard.entities.StandardEntityURN.ROAD;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Building;

import MIC.util.CollectionUtil;
import MIC.util.Convertors;
import MIC.util.RandomGenerator;


public class MIC_PFConfig {
	
	private WorldInfo worldInfo;
	private AgentInfo agentInfo;
	private PathPlanning pathPlanning;
	private ScenarioInfo scenarioInfo;

	private List<StandardEntity> policeAgents;

	private CollectionUtil collectionUtil;
	private EntityID Comm_chosenTarget = null;
	private RandomGenerator randomGen;
	private Convertors convertor;
	
	public MIC_PFConfig(WorldInfo worldInfo, AgentInfo agentInfo, PathPlanning pathPlanning, ScenarioInfo scenarioInfo)
	{
		this.worldInfo = worldInfo;
		this.agentInfo = agentInfo;
		this.pathPlanning = pathPlanning;
		this.scenarioInfo = scenarioInfo;
		this.policeAgents = new ArrayList<StandardEntity>(this.worldInfo.getEntitiesOfType(rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE));
		randomGen = new RandomGenerator();
		convertor = new Convertors(this.worldInfo);
		collectionUtil = new CollectionUtil();
	}
	
	public int getRepairCost(Road road) {
        int cost = 0;
        if(road.isBlockadesDefined()) {
            for(EntityID id : road.getBlockades())
            {
                Blockade blockade = (Blockade) this.worldInfo.getEntity(id);
                if(blockade != null && blockade.isRepairCostDefined())
                {
                    cost += blockade.getRepairCost();
                }
            }
        }
        return cost;
    }
	
	public List<EntityID> removeUselessRoads(List<EntityID> currentRoads)
	{
		List<EntityID> changedEntities = new ArrayList<EntityID>(this.worldInfo.getChanged().getChangedEntities());
		for(EntityID eid : changedEntities)
		{
			Road road = (Road) this.worldInfo.getEntity(eid);
			if( road.isBlockadesDefined() == false || road.getBlockades().isEmpty() )
				collectionUtil.removeItemFromList(currentRoads, eid);
		}
		return currentRoads;
	
	}

	public List<EntityID> selectNearbyRoads(EntityID Refuge, WorldInfo worldInfo)
	{
		List<EntityID> nearby = new ArrayList<EntityID>();
		List<EntityID> nearbyRoads = new ArrayList<EntityID>();
		
		int range = 200;
	
		while(nearbyRoads.size() < 20)
		{
			
			nearby = new ArrayList<EntityID>(worldInfo.getObjectIDsInRange(Refuge, range));
			for(EntityID ei : nearby)
			{
				StandardEntity se = worldInfo.getEntity(ei);
				if ( nearbyRoads.contains(ei) )
					continue;
				
				List<StandardEntity> roads = new ArrayList<StandardEntity>(worldInfo.getEntitiesOfType(ROAD));
				if(roads.contains( se ))
				{
					nearbyRoads.add( se.getID() );
				}
			}
			if(nearbyRoads.size() < 20)
				range += 100;	
		}
		
		return nearbyRoads;
	}

    public List<StandardEntity> detectRefugeEntrance(StandardEntity refuge, WorldInfo wi, PathPlanning pathPlanning)
    {
    	List<StandardEntity> FPRoads = new ArrayList<StandardEntity>();
    	List<EntityID> roads = selectNearbyRoads(refuge.getID() , wi);
		Map<StandardEntity, Integer> RoadDistances = new HashMap<StandardEntity, Integer>();
	
		for(EntityID road : roads)
		{
			StandardEntity standardRoad = wi.getEntity(road);
    		
			if(road != null)
			{
				int distance = pathPlanning.getResult(road, refuge.getID()).size();
    			RoadDistances.put(standardRoad, distance);
    		}
		}
		
		Map<StandardEntity, Integer> sorted = collectionUtil.MapSorter(RoadDistances, true);
		List<StandardEntity> finalList = new ArrayList<StandardEntity>(sorted.keySet());
		
		for(Iterator<StandardEntity> iterator = finalList.iterator(); iterator.hasNext();)
		{        		
			StandardEntity object = iterator.next();
			if(object != finalList.get(0) && object != finalList.get(1))
				iterator.remove();
		}
		
		FPRoads.addAll(finalList);
		return FPRoads;
    }

	public List<CommandPolice> gatherMessages(MessageManager messageManager)
	{
		List<CommunicationMessage> RawReceivedMessages = messageManager.getReceivedMessageList();
				
		List<CommandPolice> receivedMessages = new ArrayList<>();
		for( CommunicationMessage RawCommand : RawReceivedMessages)
		{
			if(RawCommand.getClass() == CommandPolice.class)
			{
				CommandPolice PoliceCommand = (CommandPolice)RawCommand;
                if (PoliceCommand.getAction() == CommandPolice.ACTION_CLEAR)
                {
                    receivedMessages.add(PoliceCommand);
                }
			}
		}
		return receivedMessages;
	}

	public List<EntityID> calcPreparate()
	{
		List<EntityID> chosenEntities = new ArrayList<EntityID>();
		for (StandardEntity e : this.worldInfo.getEntitiesOfType(REFUGE, BUILDING, GAS_STATION))
		{
            for (EntityID id : ((Building) e).getNeighbours()) 
            {
                StandardEntity neighbour = this.worldInfo.getEntity(id);
                if (neighbour instanceof Road) 
                {
                    chosenEntities.add(id);
                }
            }
        }
		return chosenEntities;
	}

	public EntityID calcCommunication(List<CommandPolice> messages)
	{
		Map<EntityID, EntityID> commandList = new HashMap<EntityID, EntityID>();
		EntityID result = null;
		
        if (messages.size() > 0 && this.Comm_chosenTarget == null)
        {
            List<EntityID> targets = new ArrayList<>(messages.size());
            for (CommandPolice cmd : messages)
            {
                targets.add(cmd.getTargetID());
            }
            
            List<EntityID> policeAgentIds = convertor.getEntityIDList(policeAgents);
            
            for (EntityID targetId : targets)
            {
            	EntityID chosenAgent = null;
                int minDist = Integer.MAX_VALUE;
                int closestAgentIndex = -1;
                int i = 0;
                for (EntityID policeId : policeAgentIds)
                {
                    EntityID agentPosition = ( (Human) this.worldInfo.getEntity(policeId)).getPosition();
                    int dist = this.worldInfo.getDistance(agentPosition, targetId);
                    if (dist < minDist) 
                    {
                        minDist = dist;
                        chosenAgent = policeId;
                        closestAgentIndex = i;
                    }
                    i++;
                }
                if (closestAgentIndex >= 0)
                {
                	commandList.put(chosenAgent, targetId);
                    collectionUtil.removeItemFromList(policeAgentIds, chosenAgent);
                }
            }
            
            if (commandList.containsKey(agentInfo.getID()))
            {
            	result = commandList.get(agentInfo.getID());
                this.Comm_chosenTarget = result;
                return result;
            }
	        
	        if (this.Comm_chosenTarget != null)
	        {
		        if (agentInfo.getPosition().equals(this.Comm_chosenTarget)) this.Comm_chosenTarget = null;
	        } 
	        else
	        {
	            result = this.Comm_chosenTarget;
	            return result;
	        }
	    }
        return null;
	}
	public static List<rescuecore2.misc.geometry.Line2D> getAllBlockadeLines(Blockade b, WorldInfo wi) {
        return GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(b.getApexes()), true);
    }
	
    public static rescuecore2.standard.entities.Area getArea(WorldInfo wi, EntityID id) {
        StandardEntity se = wi.getEntity(id);
        if (se instanceof rescuecore2.standard.entities.Area) {
            return (rescuecore2.standard.entities.Area) se;
        }
        return null;
    }

    public boolean isStuck(int x, int y, EntityID areaID, WorldInfo wi) {
        int stuckThresh = 5;
        rescuecore2.standard.entities.Area area = getArea(wi, areaID);
        if (area != null)
        {
            if (area.isBlockadesDefined())
            {
                List<EntityID> blockadeIDs = area.getBlockades();
                for (EntityID blockadeID : blockadeIDs)
                {
                    Blockade blockade = (Blockade) wi.getEntity(blockadeID);
                    if (new java.awt.geom.Area(blockade.getShape()).contains(x, y))
                    {
                        return true;
                    } 
                    else {
                        List<rescuecore2.misc.geometry.Line2D> bLines = getAllBlockadeLines(blockade ,wi);
                        Point2D humanPoint = new Point2D(x, y);
                        for (rescuecore2.misc.geometry.Line2D line : bLines) {
                            Point2D closest = GeometryTools2D.getClosestPointOnSegment(line, humanPoint);
                            if (closest != null) {
                                double dist = GeometryTools2D.getDistance(closest, humanPoint);
                                if (dist < stuckThresh) {
                                    return true;
                                }
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    
    public List<StandardEntity> getStuckedHumansNearby(EntityID agent)
    {
    	Human me = (Human) this.worldInfo.getEntity(agent);
    	int myX = (int) me.getX();
    	int myY = (int) me.getY();
    
    	List<StandardEntity> result = new ArrayList<>();
        Collection<EntityID> nearbyEntities = worldInfo.getObjectIDsInRange(myX, myY, this.scenarioInfo.getPerceptionLosMaxDistance());
        
        for (EntityID eid : nearbyEntities)
        {
            if (this.worldInfo.getEntity(eid) instanceof Human &&
            		this.worldInfo.getEntity(eid).getID().getValue() != agent.getValue())
            {
                Human h = (Human) this.worldInfo.getEntity(eid);;
                if (h.isXDefined() && h.isYDefined() && h.isPositionDefined())
                {
                    if (isStuck(h.getX(), h.getY(), h.getID(), this.worldInfo))
                    {
                        if (h.isHPDefined() && h.getHP() > 0)
                        {
                            result.add(h);
                        } 
                    }
                }
            }
        }
        return result;
    }
    
    
   public EntityID calcStuckHuman()
   {
	   List<StandardEntity> traningData = this.getStuckedHumansNearby(this.agentInfo.getID());
	   EntityID chosenHuman = null;
	   
	   int minDist = Integer.MAX_VALUE;
       for (StandardEntity h : traningData)
       {
    	   int dist = this.worldInfo.getDistance(h.getID(), this.agentInfo.getID());
       
           if (dist < minDist)
           {
        	   chosenHuman = h.getID();
               minDist = dist;
           }
       }

       if (chosenHuman != null)
          return ( (Human) this.worldInfo.getEntity(chosenHuman) ).getPosition();
       else
    	   return null;
   }
   
   
   
   public EntityID calcCommon(Set<EntityID> priorityRoads, Set<EntityID> targetAreas)
   {
	   if(priorityRoads == null || priorityRoads.isEmpty())
	   {
		   List<StandardEntity> refuges = new ArrayList<>(this.worldInfo.getEntitiesOfType(REFUGE));
		   List<StandardEntity> alternativeRoadsEntities = new ArrayList<>(
				   this.detectRefugeEntrance(refuges.get(this.randomGen.GenerateRandomNumber(0, refuges.size() - 1)), this.worldInfo, this.pathPlanning)
				   );
		   
		   List<StandardEntity> removeList = new ArrayList<StandardEntity>(alternativeRoadsEntities.size());
	       for (StandardEntity se : alternativeRoadsEntities)
	       {
	           if (!targetAreas.contains(se.getID()))
	           {
	               removeList.add(se);
	           }
	       }
	       alternativeRoadsEntities.removeAll(removeList);
	       if (alternativeRoadsEntities.size() > 0)
	       {
	           this.pathPlanning.setFrom(this.agentInfo.getPosition());
	           this.pathPlanning.setDestination(targetAreas);
	           List<EntityID> path = this.pathPlanning.calc().getResult();
	           if (path != null && path.size() > 0)
	           {
	               return path.get(path.size() - 1);
	           }
	       }
	   }
	   else if(priorityRoads != null)
	   {
		   List<EntityID> removeList = new ArrayList<EntityID>(priorityRoads.size());
	       for (EntityID eid : priorityRoads)
	       {
	           if (!targetAreas.contains(eid))
	           {
	               removeList.add(eid);
	           }
	       }
	       priorityRoads.removeAll(removeList);
	       if (priorityRoads.size() > 0)
	       {
	           this.pathPlanning.setFrom(this.agentInfo.getPosition());
	           this.pathPlanning.setDestination(targetAreas);
	           List<EntityID> path = this.pathPlanning.calc().getResult();
	           if (path != null && path.size() > 0)
	           {
	               return path.get(path.size() - 1);
	           }
	       }
	   }
	   
       return null;
   }
   private Road selectRoad(Road road1, Road road2) {
       if(road1 != null && road1.isBlockadesDefined())
       {
           if(road2 != null && road2.isBlockadesDefined())
           {
               if(road1.getID().getValue() == road2.getID().getValue()) {
                   return road1;
               }
               int cost1 = this.getRepairCost(road1);
               int cost2 = this.getRepairCost(road2);
               if(cost1 > cost2)
                   return road1;
               else if(cost1 < cost2)
                   return road2;
               else
               {
                   if((cost1 + cost2) > 0)
                   {
                       if (road1.getNeighbours().size() > road2.getNeighbours().size())
                           return road1;
                       else if (road1.getNeighbours().size() < road2.getNeighbours().size())
                           return road2;
                   }
               }
           }
           if(road1.getBlockades().size() > 0) {
               return road1;
           }
       }
       if(road2 != null && road2.isBlockadesDefined() && road2.getBlockades().size() > 0) {
           return road2;
       }
       return null;
   }
   

   private Civilian selectCivilian(Civilian civilian1, Civilian civilian2) {

   	if(civilian1 != null && civilian2 == null) return civilian1;
   	else if(civilian2 != null && civilian1 == null) return civilian2;
   	else if(civilian2 == null && civilian1 == null) return null;
   	else
       {
           if(civilian1.isDirectionDefined() && civilian2.isDamageDefined())
           {
               if (civilian1.getDamage() > civilian2.getDamage())
                   return civilian1;
               
               else if (civilian1.getDamage() < civilian2.getDamage())
                   return civilian2;
           }
           else if(civilian1.isHPDefined() && civilian2.isHPDefined())
           {
               if (civilian1.getHP() < civilian2.getHP())
                   return civilian1;
               
               else if (civilian1.getHP() > civilian2.getHP())
                   return civilian2;
           }
           else if(civilian1.isBuriednessDefined() && civilian2.isBuriednessDefined())
           {
               if (civilian1.getBuriedness() > civilian2.getBuriedness())
                   return civilian1;
               
               else if (civilian1.getBuriedness() < civilian2.getBuriedness())
                   return civilian2;
           }
       }
       return null;
   }
   

   private Building selectBuilding(Building building1, Building building2) {
       
   	if(building1 != null && building2 == null) return building1;
   	else if(building2 != null && building1 == null) return building2;
   	else if(building2 == null && building1 == null) return null;
   	else
       {
           if(building2 != null)
           {
               if(building1.isOnFire() && building2.isOnFire())
               {
                   if (building1.getFieryness() > building2.getFieryness())
                       return building2;
                   
                   else if (building1.getFieryness() < building2.getFieryness())
                       return building1;
                   
               }
               else if (!building1.isOnFire() && building2.isOnFire())
                   return building2;
           }
           return building1;
       }
       
   }

   public CommunicationMessage searchComms(MessageManager messageManager, Collection<EntityID> agentPositions) {
       Human agent = (Human) this.agentInfo.me();
       Building building = null;
       Civilian civilian = null;
       Road road = null;

       EntityID position = agent.getPosition();
       
       for(EntityID eid : this.worldInfo.getChanged().getChangedEntities())
       {
           
       	StandardEntity entity = this.worldInfo.getEntity(eid);
           
       	if(isStuck((int)agentInfo.getX(), (int)agentInfo.getY(), agentInfo.getPosition(), this.worldInfo))
       	{
   			if(this.worldInfo.getEntity(this.agentInfo.getPosition()) instanceof Road)
   				road = (Road) this.worldInfo.getEntity(this.agentInfo.getPosition());
       	}

           else if(entity instanceof Building)
           {
               if(!agentPositions.contains(eid))
               {
                   building = this.selectBuilding(building, (Building) entity);
               }
               else if(eid.getValue() == position.getValue())
               {
                   building = this.selectBuilding(building, (Building) entity);
               }
           } 
       	
           else if(entity instanceof Road)
           {
               if(!agentPositions.contains(eid))
                   road = this.selectRoad(road, (Road) entity);
                   
               else if(eid.getValue() == position.getValue())
                   road = this.selectRoad(road, (Road) entity);
           }
           else if(entity.getStandardURN() == CIVILIAN)
           {
               Civilian target = (Civilian) entity;
               if(!agentPositions.contains(target.getPosition()))
                   civilian = this.selectCivilian(civilian, target);
               
               else if(target.getPosition().getValue() == position.getValue())
                   civilian = this.selectCivilian(civilian, target);
           }
           else if(entity.getStandardURN() == BLOCKADE)
           {
               Blockade blockade = (Blockade) entity;
               if(blockade.isPositionDefined())
               {
                   StandardEntity blockadePosition = this.worldInfo.getEntity(blockade.getPosition());
                   if(blockadePosition instanceof Road)
                   {
                       Road target = (Road) blockadePosition;
                       if (!agentPositions.contains(target.getID()))
                           road = this.selectRoad(road, target);
                       else if (target.getID().getValue() == position.getValue()) 
                           road = this.selectRoad(road, target);
                   }
               }
           }
       }
       if (building != null)
    	   return (new MessageBuilding(false, building));
       return null;
   }
   
   public Collection<EntityID> fillAgentPositions(AgentInfo agentInfo)
   {
	   Collection<EntityID> result = new ArrayList<EntityID>();
       for(StandardEntity entity : this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.POLICE_FORCE, StandardEntityURN.AMBULANCE_TEAM))
       {
           Human other = (Human)entity;
           result.add(other.getPosition());
       }
       return result;
   }

   public Collection<EntityID> removeWorthlessTargets(Collection<EntityID> unsearched)
   {
	   unsearched = collectionUtil.removeItemFromList(new ArrayList<EntityID>(unsearched), this.agentInfo.getPosition());
	   for (EntityID id : agentInfo.getChanged().getChangedEntities())
	   {
           StandardEntity entity = worldInfo.getEntity(id);
           if (entity instanceof Building)
           {
        	   Building b = (Building) entity;
        	   
        	   if(
    			   (b.isFierynessDefined() && ( b.getFieryness() == 8  || b.getFieryness() > 2) ) ||
    			   (b.isBrokennessDefined() && b.getBrokenness() == 0) )
        	   {
        		   unsearched = collectionUtil.removeItemFromList(new ArrayList<EntityID>(unsearched), id);   
        	   }
           }
       }
	   
	   List<EntityID> othersPosotion = new ArrayList<EntityID>();
	   othersPosotion.addAll(this.getAgentURNPositions(StandardEntityURN.AMBULANCE_TEAM));
	   othersPosotion.addAll(this.getAgentURNPositions(StandardEntityURN.POLICE_FORCE));
	   unsearched = collectionUtil.removeIDListFromList(new ArrayList<EntityID>(unsearched), othersPosotion);
	   
	   return unsearched;
   }
   
   public List<EntityID> getAgentURNPositions(StandardEntityURN urn)
   {
	   List<EntityID> positions = new ArrayList<EntityID>();
	   List<StandardEntity> agents = new ArrayList<StandardEntity>(this.worldInfo.getEntitiesOfType(urn));
	   
	   if(urn == StandardEntityURN.AMBULANCE_TEAM)
	   {
		   for(StandardEntity entity : agents)
		   {
			   if (entity instanceof AmbulanceTeam) {
	               AmbulanceTeam currentAgent = (AmbulanceTeam) entity;
	               if ( currentAgent.isPositionDefined() && currentAgent.getPosition() != null )
	               {
	                   if ( positions.contains(currentAgent.getPosition()) == false )
	                	   positions.add(currentAgent.getPosition());
	               }
	           }
		   }
	   }
	   else if(urn == StandardEntityURN.POLICE_FORCE)
	   {
		   for(StandardEntity entity : agents)
		   {
			   if (entity instanceof PoliceForce) {
	               PoliceForce currentAgent = (PoliceForce) entity;
	               if ( currentAgent.isPositionDefined() && currentAgent.getPosition() != null )
	               {
	                   if ( positions.contains(currentAgent.getPosition()) == false )
	                	   positions.add(currentAgent.getPosition());
	               }
	           }
		   }
	   }
	   
	   return positions;
   }
   
   public Collection<EntityID> customSearchReset(Collection<EntityID> unsearched, Clustering clustering, int index)
   {
	   unsearched.clear();
	   Collection<StandardEntity> clusterEntities = new ArrayList<>();
	   index -=1 ;
	   do
	   {
		   index += 1;
		   clusterEntities = clustering.getClusterEntities(index);
	   }
	   while( clusterEntities.isEmpty() && index < clustering.getClusterNumber() );
       
	   if (clusterEntities != null && clusterEntities.size() > 0)
       {
           for (StandardEntity entity : clusterEntities) {
               if (entity instanceof Building)
               {
            	   Building b = (Building) entity;
            	   if( b.isOnFire() &&
            			   b.isBrokennessDefined() && b.getBrokenness() > 0 )
				   {
            		   unsearched.add(entity.getID());
				   }
               }
           }
       }
       return unsearched;
   }
   
   
}