package MIC.util;

import adf.agent.info.AgentInfo;
import adf.agent.info.WorldInfo;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;

import java.awt.*;
import java.awt.geom.GeneralPath;
import java.awt.geom.PathIterator;
import java.util.ArrayList;
import java.util.List;
import java.awt.geom.Area;
import java.util.Collection;
import java.util.Set;

import MIC.util.Agent.MIC_ATConfig;


public class ExtraActionsUtil {
	
	/***
	 * 
	 * Thanks a lot Okan.
	 * Copied from RoboAKUT Geometry
	 * 
	 */
	
	public static Area getClearArea(int agentX, int agentY, int targetX, int targetY,
            int clearLength, int clearRad, int padding) {
			Vector2D agentToTarget = new Vector2D(targetX - agentX, targetY - agentY);

			agentToTarget = agentToTarget.normalised().scale(clearLength+padding);

			Vector2D backAgent = (new Vector2D(agentX, agentY)).add(agentToTarget.normalised().scale(-510-2*padding));
			Line2D line = new Line2D(backAgent.getX(), backAgent.getY(),agentToTarget.getX(), agentToTarget.getY());

			Vector2D dir = agentToTarget.normalised().scale(clearRad+padding);
			Vector2D perpend1 = new Vector2D(-dir.getY(), dir.getX());
			Vector2D perpend2 = new Vector2D(dir.getY(), -dir.getX());

			rescuecore2.misc.geometry.Point2D points[] = new rescuecore2.misc.geometry.Point2D[] {
					line.getOrigin().plus(perpend1),
					line.getEndPoint().plus(perpend1),
					line.getEndPoint().plus(perpend2),
					line.getOrigin().plus(perpend2) };
			int[] xPoints = new int[points.length];
			int[] yPoints = new int[points.length];
			for (int i = 0; i < points.length; i++) {
				xPoints[i] = (int) points[i].getX();
				yPoints[i] = (int) points[i].getY();
			}	
			return new Area(new Polygon(xPoints, yPoints, points.length));
	}
	
	public static boolean isInsideArea(rescuecore2.standard.entities.Area r, int x, int y) {
        List<Line2D> areaLines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(r.getApexList()));
        if (areaLines.size() <= 0)
            return false;

        GeneralPath path = new GeneralPath();
        Line2D firstLine = areaLines.get(0);
        path.moveTo(firstLine.getOrigin().getX(), firstLine.getOrigin().getY());
        for (int i = 0; i < areaLines.size(); i++) {
            Line2D line = areaLines.get(i);
            path.lineTo(line.getEndPoint().getX(), line.getEndPoint().getY());
        }
        Area area = new Area(path);
        return area.contains(x, y);
    }
	
	public static List<Line2D> getAreaLines(Area area) {
        List<Line2D> areaLines = new ArrayList<>();
        PathIterator iter = area.getPathIterator(null);
        double prevPoint[] = null;
        while(!iter.isDone()) {
            double point[] = new double[2];
            int type = iter.currentSegment(point);
            iter.next();
            if (type != 4) {
                if (prevPoint != null) {
                    areaLines.add(new Line2D(new Point2D(prevPoint[0], prevPoint[1]), new Point2D(point[0], point[1])));
                }
            }
            prevPoint = point;
        }
        return areaLines;
    }
	
	public static Point2D getAgentPosition(AgentInfo agentInfo) {
        return new Point2D(agentInfo.getX(), agentInfo.getY());
    }
	
	public static Point2D getLongerClearPoint(int agentX, int agentY, int clearX, int clearY, int clearDist) {
        Vector2D clearVector = new Vector2D(clearX-agentX, clearY-agentY);
        clearVector = clearVector.normalised().scale(clearDist+520);
        return new Point2D(agentX+clearVector.getX(), agentY+clearVector.getY());
    }
	
	public static double calDistToClosestPointToLines(List<Line2D> lines, Point2D p) {
        double minDist = Double.MAX_VALUE;
        for (Line2D line : lines) {
            Point2D closestPoint = GeometryTools2D.getClosestPointOnSegment(line, p);
            if (closestPoint != null) {
                double dist = GeometryTools2D.getDistance(closestPoint, p);
                if (dist < minDist) {
                    minDist = dist;
                }
            }
        }
        return minDist;
    }

    public static double calDistToClosestPointToBlockade(Blockade b, int x, int y) {
        List<Line2D> blockadeLines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(b.getApexes()));
        Point2D p = new Point2D(x,y);
        return calDistToClosestPointToLines(blockadeLines, p);
    }
    
    
    public static List<Human> getStuckedHumansSeenByAgent(int agentX, int agentY, EntityID agentID, WorldInfo worldInfo, ChangeSet change, int perceptionLosDistance) {
        List<Human> humansStuckInBlockades = new ArrayList<>();
        Collection<EntityID> losEntityIDs = worldInfo.getObjectIDsInRange(agentX, agentY, perceptionLosDistance);
        for (EntityID id : losEntityIDs) {
            StandardEntity se = worldInfo.getEntity(id);
            if (se instanceof Human && !se.getID().equals(agentID)) {
                Human h = (Human)se;
                if (h.isXDefined() && h.isYDefined() && h.isPositionDefined()) {
                    if (MIC_ATConfig.isStuckInBlockade(h.getX(), h.getY(), h.getPosition(), worldInfo) ||
                            isBuildingBlocked(h.getPosition(), worldInfo)) {
                        if (h.isHPDefined() && h.getHP() > 0) {
                            humansStuckInBlockades.add(h);
                        } 
                    }
                }
            }
        }

        Set<EntityID> changedEntities = change.getChangedEntities();
        for (EntityID id : changedEntities) {
            StandardEntity se = worldInfo.getEntity(id);
            if (se instanceof Human && !se.getID().equals(agentID)) {
                Human h = (Human)se;
                if (h.isXDefined() && h.isYDefined() && h.isPositionDefined()) {
                    if (MIC_ATConfig.isStuckInBlockade(h.getX(), h.getY(), h.getPosition(), worldInfo) ||
                            isBuildingBlocked(h.getPosition(), worldInfo)) {
                        if (h.isHPDefined() && h.getHP() > 0 && !humansStuckInBlockades.contains(h)) {
                            humansStuckInBlockades.add(h);
                        } 
                    }
                }
            }
        }

        return humansStuckInBlockades;
    }
    public static boolean isBuildingBlocked(EntityID areaID, WorldInfo wi) {
        StandardEntity se = wi.getEntity(areaID);
        if (se instanceof Building) {
            Building b = (Building)se;
            List<EntityID> neighbors = b.getNeighbours();
            boolean blockadeExists = false;
            for (EntityID nID : neighbors) {
                if (wi.getEntity(nID) instanceof Road) {
                    Road r = (Road) wi.getEntity(nID);
                    if (r.isBlockadesDefined() && r.getBlockades().size() > 0 ) {
                        if (!isRoadPassable(r,b,wi)) {
                            blockadeExists = true;
                        }
                    } else {
                        return false;
                    }
                }
            }
            return blockadeExists;
        } else {
            return false;
        }
    }
    
    public static boolean isEqual(Point2D point1, Point2D point2, double thresh) {
        double dist = GeometryTools2D.getDistance(point1, point2);
        if (dist < thresh) {
            return true;
        }
        return false;
    }
    
    public static boolean isPathEqual(List<EntityID> path1, List<EntityID> path2) {
        if (path1 == null || path2 == null)
            return false;

        if (path1.size() != path2.size())
            return false;

        for (int i = 0; i < path1.size(); i++) {
            EntityID id1 = path1.get(i);
            EntityID id2 = path2.get(i);
            if (!id1.equals(id2))
                return false;
        }

        return true;
    }
    
    public static List<Line2D> getAllBlockadeLines(Blockade b, WorldInfo wi) {
        return GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(b.getApexes()), true);
    }

    public static double calDistToClosestPointToAreaLines(rescuecore2.standard.entities.Area r, int x, int y) {
        List<Line2D> areaLines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(r.getApexList()));
        Point2D p = new Point2D(x,y);
        return calDistToClosestPointToLines(areaLines, p);
    }
    
    public static Vector2D getClearVector(int agentX, int agentY, int targetX, int targetY, int clearLength, int padding) {
        Vector2D agentToTarget = new Vector2D(targetX - agentX, targetY - agentY);
        agentToTarget = agentToTarget.normalised().scale(clearLength+padding);
        Vector2D backAgent = (new Vector2D(agentX, agentY))
                .add(agentToTarget.normalised().scale(-510-2*padding));
        return new Vector2D(backAgent.getX()+agentToTarget.getX(), backAgent.getY()+agentToTarget.getY());
    }
    
    public static List<Line2D> getAllBlockadeLines(Road r, WorldInfo wi) {
        List<Line2D> blockadeLines = new ArrayList<>();
        if (r.isBlockadesDefined()) {
            List<EntityID> blockades = r.getBlockades();
            for (EntityID bID : blockades) {
                Blockade b = (Blockade)wi.getEntity(bID);
                blockadeLines.addAll(GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(b.getApexes()), true));
            }
        }
        return blockadeLines;
    }
    
    public static boolean isRoadPassable(Road road, Building b, WorldInfo wi) {
        List<Line2D> blockadeLines = getAllBlockadeLines(road, wi);
        // check all the passages between double combinations of edges
        List<Edge> roadEdges = road.getEdges();

        // find the edge between the road end the building
        Edge bEdge = null;
        for (Edge e : roadEdges) {
            if (e.isPassable() && e.getNeighbour().equals(b.getID())) {
                bEdge = e;
                break;
            }
        }

        // the direct passage between the building edge center the other edges
        // if there is any direct line between two edges which does not hit a blockade line
        // it means agent can pass
        for (Edge e : roadEdges) {
            if (e.isPassable() && !e.equals(bEdge)) {
                Line2D walkLine = new Line2D(new Point2D((e.getStartX()+e.getEndX())/2, (e.getStartY()+e.getEndY())/2),
                        new Point2D((bEdge.getStartX()+bEdge.getEndX())/2, (bEdge.getStartY()+bEdge.getEndY())/2));
                boolean blocked = false;
                for (Line2D bLine : blockadeLines) {
                    Point2D intersection = GeometryTools2D.getSegmentIntersectionPoint(walkLine, bLine);
                    if (intersection != null) {
                        blocked = true;
                        break;
                    }
                }
                if (!blocked) {
                    return true;
                }
            }
        }

        boolean isPassableFromEdgeToCenter = false;

        // checking direct passage between other edges and center
        Point2D roadCenter = new Point2D(road.getX(), road.getY());
        for (Edge e : roadEdges) {
            if (e.isPassable() && !e.equals(bEdge)) {
                Line2D walkLine = new Line2D(new Point2D((e.getStartX()+e.getEndX())/2, (e.getStartY()+e.getEndY())/2),
                        roadCenter);
                boolean blocked = false;
                for (Line2D bLine : blockadeLines) {
                    Point2D intersection = GeometryTools2D.getSegmentIntersectionPoint(walkLine, bLine);
                    if (intersection != null) {
                        blocked = true;
                        break;
                    }
                }
                if (!blocked) {
                    isPassableFromEdgeToCenter = true;
                    break;
                }
            }
        }

        if (isPassableFromEdgeToCenter) {
            // checking direct passage between building edge center to road center
            Point2D buildingEdgeCenter = new Point2D((bEdge.getStartX() + bEdge.getEndX()) / 2, (bEdge.getStartY() + bEdge.getEndY()) / 2);
            Line2D walkLine = new Line2D(roadCenter, buildingEdgeCenter);
            boolean blocked = false;
            for (Line2D bLine : blockadeLines) {
                Point2D intersection = GeometryTools2D.getSegmentIntersectionPoint(walkLine, bLine);
                if (intersection != null) {
                    blocked = true;
                    break;
                }
            }
            if (!blocked) {
                return true;
            }
        }

        return false;
    }
	
	
			
    
    
}
